import datetime as dt
from collections import defaultdict
from enum import Enum

from django_datatables_view.base_datatable_view import BaseDatatableView

from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.mixins import LoginRequiredMixin, PermissionRequiredMixin
from django.contrib.auth.models import User
from django.db import models
from django.db.models import (
    Case,
    Count,
    ExpressionWrapper,
    F,
    FloatField,
    IntegerField,
    Min,
    OuterRef,
    Q,
    Subquery,
    Sum,
    Value,
    When,
)
from django.db.models.functions import Coalesce, Concat
from django.http import HttpResponseNotFound, JsonResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.html import format_html, strip_tags
from django.utils.timezone import now
from django.views.decorators.cache import cache_page
from esi.decorators import token_required

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.evelinks import dotlan
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from app_utils.allianceauth import notify_admins
from app_utils.logging import LoggerAddTag
from app_utils.messages import messages_plus
from app_utils.views import fontawesome_modal_button_html, link_html, yesno_str

from . import __title__, helpers, tasks
from .app_settings import (
    MOONMINING_ADMIN_NOTIFICATIONS_ENABLED,
    MOONMINING_COMPLETED_EXTRACTIONS_HOURS_UNTIL_STALE,
    MOONMINING_REPROCESSING_YIELD,
    MOONMINING_USE_REPROCESS_PRICING,
    MOONMINING_VOLUME_PER_MONTH,
)
from .constants import DATE_FORMAT, DATETIME_FORMAT, EveGroupId
from .forms import MoonScanForm
from .helpers import HttpResponseUnauthorized
from .models import EveOreType, Extraction, Moon, Owner, Refinery

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class ExtractionsCategory(str, helpers.EnumToDict, Enum):
    UPCOMING = "upcoming"
    PAST = "past"


class MoonsCategory(str, helpers.EnumToDict, Enum):
    ALL = "all_moons"
    UPLOADS = "uploads"
    OURS = "our_moons"


def moon_link_html(moon: Moon) -> str:
    return format_html(
        '<a href="#" data-toggle="modal" '
        'data-target="#modalMoonDetails" '
        'title="Show details for this moon." '
        "data-ajax_url={}>"
        "{}</a>",
        reverse("moonmining:moon_details", args=[moon.pk]),
        moon.name,
    )


def extraction_ledger_button_html(extraction: Extraction) -> str:
    return fontawesome_modal_button_html(
        modal_id="modalExtractionLedger",
        fa_code="fas fa-table",
        ajax_url=reverse("moonmining:extraction_ledger", args=[extraction.pk]),
        tooltip="Extraction ledger",
    )


def moon_details_button_html(moon: Moon) -> str:
    return fontawesome_modal_button_html(
        modal_id="modalMoonDetails",
        fa_code="fas fa-moon",
        ajax_url=reverse("moonmining:moon_details", args=[moon.pk]),
        tooltip="Moon details",
    )


def extraction_details_button_html(extraction_pk: int) -> str:
    return fontawesome_modal_button_html(
        modal_id="modalExtractionDetails",
        fa_code="fas fa-hammer",
        ajax_url=reverse("moonmining:extraction_details", args=[extraction_pk]),
        tooltip="Extraction details",
    )


def default_if_none(value, default):
    """Return given default if value is None"""
    if value is None:
        return default
    return value


def default_if_false(value, default):
    """Return given default if value is False"""
    if not value:
        return default
    return value


@login_required
@permission_required("moonmining.basic_access")
def index(request):
    if request.user.has_perm("moonmining.extractions_access"):
        return redirect("moonmining:extractions")
    return redirect("moonmining:moons")


@login_required
@permission_required(["moonmining.extractions_access", "moonmining.basic_access"])
def extractions(request):
    context = {
        "page_title": "Extractions",
        "ExtractionsCategory": ExtractionsCategory.to_dict(),
        "ExtractionsStatus": Extraction.Status,
        "use_reprocess_pricing": MOONMINING_USE_REPROCESS_PRICING,
        "reprocessing_yield": MOONMINING_REPROCESSING_YIELD * 100,
        "total_volume_per_month": MOONMINING_VOLUME_PER_MONTH / 1000000,
        "stale_hours": MOONMINING_COMPLETED_EXTRACTIONS_HOURS_UNTIL_STALE,
    }
    return render(request, "moonmining/extractions.html", context)


@login_required
@permission_required(["moonmining.extractions_access", "moonmining.basic_access"])
def extractions_data(request, category):
    data = list()
    cutover_dt = now() - dt.timedelta(
        hours=MOONMINING_COMPLETED_EXTRACTIONS_HOURS_UNTIL_STALE
    )
    extractions = (
        Extraction.objects.annotate_volume()
        .selected_related_defaults()
        .select_related(
            "refinery__moon__eve_moon__eve_planet__eve_solar_system",
            "refinery__moon__eve_moon__eve_planet__eve_solar_system__eve_constellation",
            "refinery__moon__eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
        )
    )
    if category == ExtractionsCategory.UPCOMING:
        extractions = extractions.filter(auto_fracture_at__gte=cutover_dt).exclude(
            status=Extraction.Status.CANCELED
        )
    elif category == ExtractionsCategory.PAST:
        extractions = extractions.filter(
            auto_fracture_at__lt=cutover_dt
        ) | extractions.filter(status=Extraction.Status.CANCELED)
    else:
        extractions = Extraction.objects.none()
    can_see_ledger = request.user.has_perm("moonmining.view_moon_ledgers")
    for extraction in extractions:
        corporation_name = extraction.refinery.owner.name
        alliance_name = extraction.refinery.owner.alliance_name
        moon = extraction.refinery.moon
        moon_name = str(moon)
        refinery_name = str(extraction.refinery.name)
        solar_system = moon.eve_moon.eve_planet.eve_solar_system
        constellation = region = solar_system.eve_constellation
        region = constellation.eve_region
        location = format_html(
            "{}<br><i>{}</i>",
            link_html(dotlan.solar_system_url(solar_system.name), moon_name),
            region.name,
        )
        if (
            extraction.status == Extraction.Status.COMPLETED
            and extraction.ledger.exists()
        ):
            mined_value = extraction.ledger.aggregate(Sum(F("total_price")))[
                "total_price__sum"
            ]
            actions_html = (
                extraction_ledger_button_html(extraction) + "&nbsp;"
                if can_see_ledger
                else ""
            )
        else:
            actions_html = ""
            mined_value = None
        actions_html += extraction_details_button_html(extraction.pk)
        actions_html += "&nbsp;" + moon_details_button_html(extraction.refinery.moon)
        status_html = format_html(
            "{}<br>{}",
            extraction.chunk_arrival_at.strftime(DATETIME_FORMAT),
            extraction.status_enum.bootstrap_tag_html,
        )
        data.append(
            {
                "id": extraction.pk,
                "chunk_arrival_at": {
                    "display": status_html,
                    "sort": extraction.chunk_arrival_at,
                },
                "refinery": {
                    "display": extraction.refinery.name_html(),
                    "sort": refinery_name,
                },
                "location": {
                    "display": location,
                    "sort": moon_name,
                },
                "labels": moon.labels_html(),
                "volume": extraction.volume,
                "value": extraction.value if extraction.value else None,
                "mined_value": mined_value,
                "details": actions_html,
                "corporation_name": corporation_name,
                "alliance_name": alliance_name,
                "moon_name": moon_name,
                "region_name": region.name,
                "constellation_name": constellation.name,
                "rarity_class": moon.get_rarity_class_display(),
                "is_jackpot_str": yesno_str(extraction.is_jackpot),
                "is_ready": extraction.chunk_arrival_at <= now(),
                "status": extraction.status,
                "status_str": Extraction.Status(extraction.status).label,
            }
        )
    return JsonResponse(data, safe=False)


@login_required
@permission_required(["moonmining.extractions_access", "moonmining.basic_access"])
def extraction_details(request, extraction_pk: int):
    try:
        extraction = (
            Extraction.objects.annotate_volume()
            .select_related(
                "refinery",
                "refinery__moon",
                "refinery__moon__eve_moon",
                "refinery__moon__eve_moon__eve_planet__eve_solar_system",
                "refinery__moon__eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
                "canceled_by",
                "fractured_by",
                "started_by",
            )
            .get(pk=extraction_pk)
        )
    except Extraction.DoesNotExist:
        return HttpResponseNotFound()
    context = {
        "page_title": (
            f"{extraction.refinery.moon} "
            f"| {extraction.chunk_arrival_at.strftime(DATE_FORMAT)}"
        ),
        "extraction": extraction,
    }
    if request.GET.get("new_page"):
        context["title"] = "Extraction"
        context["content_file"] = "moonmining/partials/extraction_details.html"
        return render(request, "moonmining/_generic_modal_page.html", context)
    else:
        return render(request, "moonmining/modals/extraction_details.html", context)


@login_required
@permission_required(
    [
        "moonmining.extractions_access",
        "moonmining.basic_access",
        "moonmining.view_moon_ledgers",
    ]
)
def extraction_ledger(request, extraction_pk: int):
    try:
        extraction = (
            Extraction.objects.all()
            .select_related(
                "refinery",
                "refinery__moon",
                "refinery__moon__eve_moon__eve_planet__eve_solar_system",
                "refinery__moon__eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
            )
            .get(pk=extraction_pk)
        )
    except Extraction.DoesNotExist:
        return HttpResponseNotFound()
    ledger = extraction.ledger.select_related(
        "character", "corporation", "user__profile__main_character", "ore_type"
    )
    total_value = ledger.aggregate(Sum(F("total_price")))["total_price__sum"]
    total_volume = ledger.aggregate(Sum(F("total_volume")))["total_volume__sum"]
    sum_price = ExpressionWrapper(
        F("quantity") * Coalesce(F("unit_price"), 0), output_field=FloatField()
    )
    sum_volume = ExpressionWrapper(
        F("quantity") * F("ore_type__volume"), output_field=IntegerField()
    )
    character_totals = (
        ledger.values(
            character_name=F("character__name"),
            main_name=F("user__profile__main_character__character_name"),
            corporation_name=F("user__profile__main_character__corporation_name"),
        )
        .annotate(character_total_price=Sum(sum_price, distinct=True))
        .annotate(character_total_volume=Sum(sum_volume, distinct=True))
        .annotate(
            character_percent_value=ExpressionWrapper(
                F("character_total_price") / Value(total_value) * Value(100),
                output_field=IntegerField(),
            )
        )
        .annotate(
            character_percent_volume=F("character_total_volume")
            / Value(total_volume)
            * Value(100)
        )
    )
    context = {
        "page_title": (
            f"{extraction.refinery.moon} "
            f"| {extraction.chunk_arrival_at.strftime(DATE_FORMAT)}"
        ),
        "extraction": extraction,
        "total_value": total_value,
        "total_volume": total_volume,
        "ledger": ledger,
        "character_totals": character_totals,
    }
    if request.GET.get("new_page"):
        context["title"] = "Extraction Ledger"
        context["content_file"] = "moonmining/partials/extraction_ledger.html"
        return render(request, "moonmining/_generic_modal_page.html", context)
    return render(request, "moonmining/modals/extraction_ledger.html", context)


@login_required()
@permission_required("moonmining.basic_access")
def moons(request):
    context = {
        "page_title": "Moons",
        "MoonsCategory": MoonsCategory.to_dict(),
        "use_reprocess_pricing": MOONMINING_USE_REPROCESS_PRICING,
        "reprocessing_yield": MOONMINING_REPROCESSING_YIELD * 100,
        "total_volume_per_month": MOONMINING_VOLUME_PER_MONTH / 1000000,
    }
    return render(request, "moonmining/moons.html", context)


class MoonListJson(PermissionRequiredMixin, LoginRequiredMixin, BaseDatatableView):
    model = Moon
    permission_required = "moonmining.basic_access"
    columns = [
        "id",
        "moon_name",
        "rarity_class_str",
        "refinery",
        "labels",
        "solar_system_link",
        "location_html",
        "region_name",
        "constellation_name",
        "value",
        "details",
        "has_refinery_str",
        "has_extraction_str",
        "solar_system_name",
        "corporation_name",
        "alliance_name",
        "has_refinery",
        "label_name",
    ]

    # define column names that will be used in sorting
    # order is important and should be same as order of columns
    # displayed by datatables. For non sortable columns use empty
    # value like ''
    order_columns = [
        "pk",
        "name",
        "refinery__eve_solar_system__name",
        "refinery__name",
        "",
        "value",
        "",
        # hidden columns below
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
    ]

    def get_initial_queryset(self) -> models.QuerySet:
        return self.initial_queryset(
            category=self.kwargs["category"], user=self.request.user
        )

    @classmethod
    def initial_queryset(cls, category: str, user: User) -> models.QuerySet:
        current_extraction_qs = Extraction.objects.filter(
            refinery__moon=OuterRef("pk"),
            status__in=[Extraction.Status.STARTED, Extraction.Status.READY],
        )
        moon_query = (
            Moon.objects.selected_related_defaults()
            .annotate(extraction_pk=Subquery(current_extraction_qs.values("pk")[:1]))
            .annotate(
                has_refinery=Case(
                    When(refinery__isnull=True, then=Value(False)), default=Value(True)
                )
            )
            .annotate(
                has_refinery_str=Case(
                    When(has_refinery=False, then=Value("no")), default=Value("yes")
                )
            )
            .annotate(
                has_extraction=Case(
                    When(extraction_pk__isnull=True, then=Value(False)),
                    default=Value(True),
                )
            )
            .annotate(
                has_extraction_str=Case(
                    When(has_extraction=False, then=Value("no")), default=Value("yes")
                )
            )
            .annotate(
                rarity_class_str=Concat(
                    Value("R"), F("rarity_class"), output_field=models.CharField()
                )
            )
        )
        if (
            category == MoonsCategory.ALL
            and user.has_perm("moonmining.extractions_access")
            and user.has_perm("moonmining.view_all_moons")
        ):
            pass
        elif category == MoonsCategory.OURS and user.has_perm(
            "moonmining.extractions_access"
        ):
            moon_query = moon_query.filter(refinery__isnull=False)
        elif category == MoonsCategory.UPLOADS and user.has_perm(
            "moonmining.upload_moon_scan"
        ):
            moon_query = moon_query.filter(products_updated_by=user)
        else:
            moon_query = Moon.objects.none()
        return moon_query

    def filter_queryset(self, qs) -> models.QuerySet:
        """use parameters passed in GET request to filter queryset"""

        qs = self._apply_search_filter(
            qs, 7, "eve_moon__eve_planet__eve_solar_system__name"
        )
        qs = self._apply_search_filter(qs, 8, "has_refinery_str")
        qs = self._apply_search_filter(
            qs, 9, "refinery__owner__corporation__corporation_name"
        )
        qs = self._apply_search_filter(
            qs, 10, "refinery__owner__corporation__alliance__alliance_name"
        )
        qs = self._apply_search_filter(qs, 11, "rarity_class_str")
        qs = self._apply_search_filter(qs, 12, "has_extraction_str")
        qs = self._apply_search_filter(
            qs, 13, "eve_moon__eve_planet__eve_solar_system__eve_constellation__name"
        )
        qs = self._apply_search_filter(qs, 14, "label__name")
        qs = self._apply_search_filter(
            qs,
            15,
            "eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region__name",
        )

        search = self.request.GET.get("search[value]", None)
        if search:
            qs = qs.filter(
                Q(eve_moon__name__istartswith=search)
                | Q(refinery__name__istartswith=search)
            )
        return qs

        qs = self._apply_search_filter(qs, 4, "user__profile__state__name")
        qs = self._apply_search_filter(qs, 6, "character__alliance_name")
        qs = self._apply_search_filter(qs, 7, "character__corporation_name")
        qs = self._apply_search_filter(
            qs, 8, "user__profile__main_character__alliance_name"
        )
        qs = self._apply_search_filter(
            qs, 9, "user__profile__main_character__corporation_name"
        )
        qs = self._apply_search_filter(
            qs, 10, "user__profile__main_character__character_name"
        )
        qs = self._apply_search_filter(qs, 11, "unregistered")

        return qs

    def _apply_search_filter(self, qs, column_num, field) -> models.QuerySet:
        my_filter = self.request.GET.get(f"columns[{column_num}][search][value]", None)
        if my_filter:
            if self.request.GET.get(f"columns[{column_num}][search][regex]", False):
                kwargs = {f"{field}__iregex": my_filter}
            else:
                kwargs = {f"{field}__istartswith": my_filter}
            return qs.filter(**kwargs)
        return qs

    def render_column(self, row, column) -> str:
        if column == "id":
            return row.pk
        if column == "moon_name":
            return row.name
        result = self._render_location(row, column)
        if result:
            return result
        if column == "labels":
            return row.labels_html()
        if column == "label_name":
            return row.label.name if row.label else ""
        if column == "details":
            return self._render_details(row)
        result = self._render_refinery(row, column)
        if result:
            return result
        return super().render_column(row, column)

    def _render_location(self, row, column):
        solar_system = row.eve_moon.eve_planet.eve_solar_system
        if solar_system.is_high_sec:
            sec_class = "text-high-sec"
        elif solar_system.is_low_sec:
            sec_class = "text-low-sec"
        else:
            sec_class = "text-null-sec"
        solar_system_link = format_html(
            '{}&nbsp;<span class="{}">{}</span>',
            link_html(dotlan.solar_system_url(solar_system.name), solar_system.name),
            sec_class,
            round(solar_system.security_status, 1),
        )
        constellation = row.eve_moon.eve_planet.eve_solar_system.eve_constellation
        region = constellation.eve_region
        location_html = format_html(
            "{}<br><em>{}</em>", constellation.name, region.name
        )
        if column == "solar_system_name":
            return solar_system.name
        if column == "solar_system_link":
            return solar_system_link
        if column == "location_html":
            return location_html
        if column == "region_name":
            return region.name
        if column == "constellation_name":
            return constellation.name
        return None

    def _render_details(self, row):
        has_details_access = self.request.user.has_perm(
            "moonmining.extractions_access"
        ) or self.request.user.has_perm("moonmining.view_all_moons")
        if has_details_access:
            details_html = (
                extraction_details_button_html(row.extraction_pk) + " "
                if row.extraction_pk
                else ""
            )
            details_html += moon_details_button_html(row)
            return details_html
        return ""

    def _render_refinery(self, row, column):
        if row.has_refinery:
            refinery = row.refinery
            refinery_html = refinery.name_html()
            refinery_name = refinery.name
            corporation_name = refinery.owner.name
            alliance_name = refinery.owner.alliance_name
        else:
            refinery_html = "?"
            refinery_name = ""
            corporation_name = alliance_name = ""
        if column == "corporation_name":
            return corporation_name
        if column == "alliance_name":
            return alliance_name
        if column == "refinery":
            return {"display": refinery_html, "sort": refinery_name}
        return None


@login_required
@permission_required("moonmining.basic_access")
def moons_fdd_data(request, category) -> JsonResponse:
    """Provide lists for drop down fields."""
    qs = MoonListJson.initial_queryset(category=category, user=request.user)
    columns = request.GET.get("columns")
    result = dict()
    if columns:
        for column in columns.split(","):
            if column == "alliance_name":
                options = qs.exclude(
                    refinery__owner__corporation__alliance__isnull=True,
                ).values_list(
                    "refinery__owner__corporation__alliance__alliance_name", flat=True
                )
            elif column == "corporation_name":
                options = qs.exclude(refinery__isnull=True).values_list(
                    "refinery__owner__corporation__corporation_name", flat=True
                )
            elif column == "region_name":
                options = qs.values_list(
                    "eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region__name",
                    flat=True,
                )
            elif column == "constellation_name":
                options = qs.values_list(
                    "eve_moon__eve_planet__eve_solar_system__eve_constellation__name",
                    flat=True,
                )
            elif column == "solar_system_name":
                options = qs.values_list(
                    "eve_moon__eve_planet__eve_solar_system__name",
                    flat=True,
                )
            elif column == "rarity_class_str":
                options = qs.values_list("rarity_class_str", flat=True)
            elif column == "label_name":
                options = qs.exclude(label__isnull=True).values_list(
                    "label__name", flat=True
                )
            elif column == "has_refinery_str":
                options = qs.values_list("has_refinery_str", flat=True)
            elif column == "has_extraction_str":
                options = qs.values_list("has_extraction_str", flat=True)
            else:
                options = [f"** ERROR: Invalid column name '{column}' **"]
            result[column] = sorted(list(set(options)), key=str.casefold)
    return JsonResponse(result, safe=False)


@login_required
@permission_required("moonmining.basic_access")
def moon_details(request, moon_pk: int):
    try:
        moon = Moon.objects.selected_related_defaults().get(pk=moon_pk)
    except Moon.DoesNotExist:
        return HttpResponseNotFound()
    if not request.user.has_perm("moonmining.extractions_access"):
        return HttpResponseUnauthorized()
    context = {
        "page_title": moon.name,
        "moon": moon,
        "use_reprocess_pricing": MOONMINING_USE_REPROCESS_PRICING,
        "reprocessing_yield": MOONMINING_REPROCESSING_YIELD * 100,
        "total_volume_per_month": MOONMINING_VOLUME_PER_MONTH / 1000000,
    }
    if request.GET.get("new_page"):
        context["title"] = "Moon"
        context["content_file"] = "moonmining/partials/moon_details.html"
        return render(request, "moonmining/_generic_modal_page.html", context)
    return render(request, "moonmining/modals/moon_details.html", context)


@permission_required(["moonmining.add_refinery_owner", "moonmining.basic_access"])
@token_required(scopes=Owner.esi_scopes())
@login_required
def add_owner(request, token):
    try:
        character_ownership = request.user.character_ownerships.select_related(
            "character"
        ).get(character__character_id=token.character_id)
    except CharacterOwnership.DoesNotExist:
        return HttpResponseNotFound()
    try:
        corporation = EveCorporationInfo.objects.get(
            corporation_id=character_ownership.character.corporation_id
        )
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(
            corp_id=character_ownership.character.corporation_id
        )
        corporation.save()

    owner, _ = Owner.objects.update_or_create(
        corporation=corporation,
        defaults={"character_ownership": character_ownership},
    )
    tasks.update_owner.delay(owner.pk)
    messages_plus.success(request, f"Update of refineres started for {owner}.")
    if MOONMINING_ADMIN_NOTIFICATIONS_ENABLED:
        notify_admins(
            message=("%(corporation)s was added as new owner by %(user)s.")
            % {"corporation": owner, "user": request.user},
            title=f"{__title__}: Owner added: {owner}",
        )
    return redirect("moonmining:index")


@permission_required(["moonmining.basic_access", "moonmining.upload_moon_scan"])
@login_required()
def upload_survey(request):
    context = {"page_title": "Upload Moon Surveys"}
    if request.method == "POST":
        form = MoonScanForm(request.POST)
        if form.is_valid():
            scans = request.POST["scan"]
            tasks.process_survey_input.delay(scans, request.user.pk)
            messages_plus.success(
                request,
                (
                    "Your scan has been submitted for processing. You will"
                    "receive a notification once processing is complete."
                ),
            )
        else:
            messages_plus.error(
                request,
                (
                    "Oh No! Something went wrong with your moon scan submission. "
                    "Please try again."
                ),
            )
        return redirect("moonmining:moons")
    return render(request, "moonmining/modals/upload_survey.html", context=context)


def previous_month(obj: dt.datetime) -> dt.datetime:
    first = obj.replace(day=1)
    return first - dt.timedelta(days=1)


@login_required()
@permission_required(["moonmining.basic_access", "moonmining.reports_access"])
def reports(request):
    month_minus_1 = previous_month(now())
    month_minus_2 = previous_month(month_minus_1)
    month_minus_3 = previous_month(month_minus_2)
    month_format = "%b '%y"
    if (
        Refinery.objects.filter(
            owner__is_enabled=True, ledger_last_update_at__isnull=False
        )
        .exclude(ledger_last_update_ok=True)
        .exists()
    ):
        ledger_last_updated = None
    else:
        try:
            ledger_last_updated = Refinery.objects.filter(
                owner__is_enabled=True
            ).aggregate(Min("ledger_last_update_at"))["ledger_last_update_at__min"]
        except KeyError:
            ledger_last_updated = None
    context = {
        "page_title": "Reports",
        "use_reprocess_pricing": MOONMINING_USE_REPROCESS_PRICING,
        "reprocessing_yield": MOONMINING_REPROCESSING_YIELD * 100,
        "total_volume_per_month": MOONMINING_VOLUME_PER_MONTH / 1000000,
        "month_minus_3": month_minus_3.strftime(month_format),
        "month_minus_2": month_minus_2.strftime(month_format),
        "month_minus_1": month_minus_1.strftime(month_format),
        "month_current": now().strftime(month_format),
        "ledger_last_updated": ledger_last_updated,
    }
    return render(request, "moonmining/reports.html", context)


@login_required()
@permission_required(["moonmining.basic_access", "moonmining.reports_access"])
def report_owned_value_data(request):
    moon_query = Moon.objects.select_related(
        "eve_moon",
        "eve_moon__eve_planet__eve_solar_system",
        "eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
        "refinery",
        "refinery__owner",
        "refinery__owner__corporation",
        "refinery__owner__corporation__alliance",
    ).filter(refinery__isnull=False)
    corporation_moons = defaultdict(lambda: {"moons": list(), "total": 0})
    for moon in moon_query.order_by("eve_moon__name"):
        corporation_name = moon.refinery.owner.name
        corporation_moons[corporation_name]["moons"].append(moon)
        corporation_moons[corporation_name]["total"] += default_if_none(moon.value, 0)

    moon_ranks = {
        moon_pk: rank
        for rank, moon_pk in enumerate(
            moon_query.filter(value__isnull=False)
            .order_by("-value")
            .values_list("pk", flat=True)
        )
    }
    grand_total = sum(
        [corporation["total"] for corporation in corporation_moons.values()]
    )
    data = list()
    for corporation_name, details in corporation_moons.items():
        corporation = f"{corporation_name} ({len(details['moons'])})"
        counter = 0
        for moon in details["moons"]:
            grand_total_percent = (
                default_if_none(moon.value, 0) / grand_total * 100
                if grand_total > 0
                else None
            )
            rank = moon_ranks[moon.pk] + 1 if moon.pk in moon_ranks else None
            data.append(
                {
                    "corporation": corporation,
                    "moon": {"display": moon_link_html(moon), "sort": counter},
                    "region": moon.region().name,
                    "rarity_class": moon.rarity_tag_html,
                    "value": moon.value,
                    "rank": rank,
                    "total": None,
                    "is_total": False,
                    "grand_total_percent": grand_total_percent,
                }
            )
            counter += 1
        data.append(
            {
                "corporation": corporation,
                "moon": {"display": "TOTAL", "sort": counter},
                "region": None,
                "rarity_class": None,
                "value": None,
                "rank": None,
                "total": details["total"],
                "is_total": True,
                "grand_total_percent": None,
            }
        )
    return JsonResponse(data, safe=False)


@login_required()
@permission_required(["moonmining.basic_access", "moonmining.reports_access"])
def report_user_mining_data(request):
    sum_volume = ExpressionWrapper(
        F("mining_ledger__quantity") * F("mining_ledger__ore_type__volume"),
        output_field=FloatField(),
    )
    sum_price = ExpressionWrapper(
        F("mining_ledger__quantity")
        * Coalesce(F("mining_ledger__ore_type__extras__current_price"), 0),
        output_field=FloatField(),
    )
    today = now()
    months_1 = today.replace(day=1) - dt.timedelta(days=1)
    months_2 = months_1.replace(day=1) - dt.timedelta(days=1)
    months_3 = months_2.replace(day=1) - dt.timedelta(days=1)
    users_mining_totals = (
        User.objects.filter(profile__main_character__isnull=False)
        .select_related("profile__main_character", "profile__state")
        .annotate(
            volume_month_0=Sum(
                sum_volume,
                filter=Q(
                    mining_ledger__day__month=today.month,
                    mining_ledger__day__year=today.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            volume_month_1=Sum(
                sum_volume,
                filter=Q(
                    mining_ledger__day__month=months_1.month,
                    mining_ledger__day__year=months_1.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            volume_month_2=Sum(
                sum_volume,
                filter=Q(
                    mining_ledger__day__month=months_2.month,
                    mining_ledger__day__year=months_2.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            volume_month_3=Sum(
                sum_volume,
                filter=Q(
                    mining_ledger__day__month=months_3.month,
                    mining_ledger__day__year=months_3.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            price_month_0=Sum(
                sum_price,
                filter=Q(
                    mining_ledger__day__month=today.month,
                    mining_ledger__day__year=today.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            price_month_1=Sum(
                sum_price,
                filter=Q(
                    mining_ledger__day__month=months_1.month,
                    mining_ledger__day__year=months_1.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            price_month_2=Sum(
                sum_price,
                filter=Q(
                    mining_ledger__day__month=months_2.month,
                    mining_ledger__day__year=months_2.year,
                ),
                distinct=True,
            )
        )
        .annotate(
            price_month_3=Sum(
                sum_price,
                filter=Q(
                    mining_ledger__day__month=months_3.month,
                    mining_ledger__day__year=months_3.year,
                ),
                distinct=True,
            )
        )
    )
    data = list()
    for user in users_mining_totals:
        corporation_name = user.profile.main_character.corporation_name
        if user.profile.main_character.alliance_ticker:
            corporation_name += f" [{user.profile.main_character.alliance_ticker}]"
        if any(
            [
                user.volume_month_0,
                user.volume_month_1,
                user.volume_month_2,
                user.volume_month_3,
            ]
        ):
            data.append(
                {
                    "id": user.id,
                    "name": str(user.profile.main_character),
                    "corporation": corporation_name,
                    "state": str(user.profile.state),
                    "volume_month_0": default_if_false(user.volume_month_0, 0),
                    "volume_month_1": default_if_false(user.volume_month_1, 0),
                    "volume_month_2": default_if_false(user.volume_month_2, 0),
                    "volume_month_3": default_if_false(user.volume_month_3, 0),
                    "price_month_0": default_if_false(user.price_month_0, 0),
                    "price_month_1": default_if_false(user.price_month_1, 0),
                    "price_month_2": default_if_false(user.price_month_2, 0),
                    "price_month_3": default_if_false(user.price_month_3, 0),
                }
            )
    return JsonResponse(data, safe=False)


@login_required()
@permission_required(["moonmining.basic_access", "moonmining.reports_access"])
def report_user_uploaded_data(request) -> JsonResponse:
    data = list(
        Moon.objects.values(
            name=F("products_updated_by__profile__main_character__character_name"),
            corporation=F(
                "products_updated_by__profile__main_character__corporation_name"
            ),
            state=F("products_updated_by__profile__state__name"),
        ).annotate(num_moons=Count("eve_moon_id"))
    )
    for row in data:
        if row["name"] is None:
            row["name"] = "?"
        if row["corporation"] is None:
            row["corporation"] = "?"
        if row["state"] is None:
            row["state"] = "?"
    return JsonResponse(data, safe=False)


@login_required()
@permission_required(["moonmining.basic_access", "moonmining.reports_access"])
def report_ore_prices_data(request) -> JsonResponse:
    moon_ore_group_ids = [
        EveGroupId.UNCOMMON_MOON_ASTEROIDS,
        EveGroupId.UBIQUITOUS_MOON_ASTEROIDS,
        EveGroupId.EXCEPTIONAL_MOON_ASTEROIDS,
        EveGroupId.COMMON_MOON_ASTEROIDS,
        EveGroupId.RARE_MOON_ASTEROIDS,
    ]
    qs = (
        EveOreType.objects.filter(
            eve_group_id__in=moon_ore_group_ids,
            published=True,
            extras__isnull=False,
            extras__current_price__isnull=False,
        )
        .exclude(name__icontains=" ")
        .select_related("eve_group", "extras")
    )
    data = [
        {
            "id": obj.id,
            "name": obj.name,
            "description": strip_tags(obj.description),
            "price": obj.extras.current_price,
            "group": obj.eve_group.name,
            "rarity_html": {
                "display": obj.rarity_class.bootstrap_tag_html,
                "sort": obj.rarity_class.label,
            },
            "rarity_str": obj.rarity_class.label,
        }
        for obj in qs
    ]
    return JsonResponse(data, safe=False)


@cache_page(3600)
def modal_loader_body(request):
    """Draw the loader body. Useful for showing a spinner while loading a modal."""
    return render(request, "moonmining/modals/loader_body.html")


def tests(request):
    """Render page with JS tests."""
    return render(request, "moonmining/tests.html")



import datetime as dt

from django.test import TestCase
from django.utils.timezone import now

from ..core import CalculatedExtraction, CalculatedExtractionProduct


class TestCalculatedExtractionProduct(TestCase):
    def test_should_create_list(self):
        # given
        ores = {"1": 100, "2": 200}  # keys are string because they come from JSON
        # when
        lst = CalculatedExtractionProduct.create_list_from_dict(ores)
        # then
        self.assertListEqual(
            lst,
            [
                CalculatedExtractionProduct(ore_type_id=1, volume=100),
                CalculatedExtractionProduct(ore_type_id=2, volume=200),
            ],
        )

    def test_should_calculate_share(self):
        # given
        product = CalculatedExtractionProduct(ore_type_id=1, volume=400)
        # when/then
        self.assertEqual(product.calculated_share(5, 100), 0.8)


class TestCalculatedExtraction(TestCase):
    def test_should_calculate_duration(self):
        # given
        base_dt = now()
        extraction = CalculatedExtraction(
            refinery_id=1,
            status=CalculatedExtraction.Status.STARTED,
            started_at=base_dt,
            chunk_arrival_at=base_dt + dt.timedelta(days=10),
        )
        # when/then
        self.assertEqual(extraction.duration, dt.timedelta(days=10))

    def test_should_raise_exception_when_trying_to_calculate_duration_1(self):
        # given
        base_dt = now()
        extraction = CalculatedExtraction(
            refinery_id=1,
            status=CalculatedExtraction.Status.STARTED,
            started_at=base_dt,
        )
        # when/then
        with self.assertRaises(ValueError):
            extraction.duration

    def test_should_raise_exception_when_trying_to_calculate_duration_2(self):
        # given
        base_dt = now()
        extraction = CalculatedExtraction(
            refinery_id=1,
            status=CalculatedExtraction.Status.STARTED,
            chunk_arrival_at=base_dt + dt.timedelta(days=10),
        )
        # when/then
        with self.assertRaises(ValueError):
            extraction.duration

    def test_should_calculate_total_volume(self):
        # given
        extraction = CalculatedExtraction(
            refinery_id=1, status=CalculatedExtraction.Status.STARTED
        )
        ores = {"45506": 10000, "46676": 20000}
        extraction.products = CalculatedExtractionProduct.create_list_from_dict(ores)
        # when/then
        self.assertEqual(extraction.total_volume(), 30000)

    def test_should_calculate_total_volume_without_products(self):
        # given
        extraction = CalculatedExtraction(
            refinery_id=1, status=CalculatedExtraction.Status.STARTED
        )
        # when/then
        self.assertEqual(extraction.total_volume(), 0)

    def test_should_calculate_moon_products_estimate(self):
        # given
        base_time = now()
        extraction = CalculatedExtraction(
            refinery_id=1,
            status=CalculatedExtraction.Status.STARTED,
            started_at=base_time,
            chunk_arrival_at=base_time + dt.timedelta(days=20),
        )
        ores = {"45506": 7_683_200, "46676": 9_604_000}
        extraction.products = CalculatedExtractionProduct.create_list_from_dict(ores)
        # when
        products = extraction.moon_products_estimated(960_400)
        # then
        self.assertAlmostEqual(products[0].amount, 0.4, places=2)
        self.assertAlmostEqual(products[1].amount, 0.5, places=2)

    def test_should_calculate_moon_products_estimate_with_overflow(self):
        # given
        base_time = now()
        extraction = CalculatedExtraction(
            refinery_id=1,
            status=CalculatedExtraction.Status.STARTED,
            started_at=base_time,
            chunk_arrival_at=base_time + dt.timedelta(days=20),
        )
        ores = {"45506": 10_564_400, "46676": 12_677_280}
        extraction.products = CalculatedExtractionProduct.create_list_from_dict(ores)
        # when
        products = extraction.moon_products_estimated(960_400)
        # then
        self.assertAlmostEqual(products[0].amount, 0.45, places=2)
        self.assertAlmostEqual(products[1].amount, 0.55, places=2)



# flake8: noqa
"""scripts generates moons for load testing

This script can be executed directly from shell.
"""

import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

import django
from django.apps import apps

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

"""SCRIPT"""
import datetime as dt
import json
import random
from pathlib import Path

from django.contrib.auth.models import User
from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveMarketPrice, EveMoon, EveType

from allianceauth.eveonline.models import EveCorporationInfo

from moonmining.app_settings import MOONMINING_VOLUME_PER_MONTH
from moonmining.models import (
    EveOreType,
    Extraction,
    ExtractionProduct,
    MiningLedgerRecord,
    Moon,
    MoonProduct,
    Owner,
    Refinery,
)

MAX_MOONS = 5
MAX_REFINERIES = 2
MAX_MINERS = 3
MAX_MINING_MONTHS = 3
MAX_MINING_ALTS_PER_MINER = 3
MAX_MINING_DAYS = 10
DUMMY_CORPORATION_ID = 1000127  # Guristas
DUMMY_CHARACTER_ID = 3019491  # Guristas CEO


def random_percentages(parts) -> list:
    percentages = []
    total = 0
    for _ in range(parts - 1):
        part = random.randint(0, 100 - total)
        percentages.append(part)
        total += part
    percentages.append(100 - total)
    return percentages


def generate_extraction(refinery, chunk_arrival_at, started_by, status):
    extraction = Extraction.objects.create(
        refinery=refinery,
        chunk_arrival_at=chunk_arrival_at,
        auto_fracture_at=chunk_arrival_at + dt.timedelta(hours=4),
        started_by=started_by,
        started_at=chunk_arrival_at - dt.timedelta(days=14),
        status=status,
    )
    for product in moon.products.all():
        ExtractionProduct.objects.create(
            extraction=extraction,
            ore_type=product.ore_type,
            volume=MOONMINING_VOLUME_PER_MONTH * product.amount,
        )
    return extraction


data_path = Path(__file__).parent / "generate_moons.json"
with data_path.open("r", encoding="utf-8") as fp:
    data = json.load(fp)
moon_ids = [int(obj["moon_id"]) for obj in data["moons"]]
ore_type_ids = [int(obj["type_id"]) for obj in data["ore_type_ids"]]

print(f"Generating {MAX_MOONS} moons...")
random_user = User.objects.order_by("?").first()
my_moons = list()
for moon_id in random.sample(moon_ids, k=MAX_MOONS):
    print(f"Creating moon {moon_id}")
    eve_moon, _ = EveMoon.objects.get_or_create_esi(id=moon_id)
    moon, created = Moon.objects.get_or_create(
        eve_moon=eve_moon,
        defaults={
            "products_updated_at": now(),
            "products_updated_by": random_user,
        },
    )
    my_moons.append(moon)
    if created:
        percentages = random_percentages(4)
        for ore_type_id in random.sample(ore_type_ids, k=4):
            ore_type, _ = EveOreType.objects.get_or_create_esi(id=ore_type_id)
            MoonProduct.objects.create(
                moon=moon, ore_type=ore_type, amount=percentages.pop()
            )
        moon.update_calculated_properties()
print(f"Generating {MAX_REFINERIES} refineries...")
try:
    corporation = EveCorporationInfo.objects.get(corporation_id=DUMMY_CORPORATION_ID)
except EveCorporationInfo.DoesNotExist:
    corporation = EveCorporationInfo.objects.create_corporation(
        corp_id=DUMMY_CORPORATION_ID
    )
owner, _ = Owner.objects.get_or_create(corporation=corporation)
Refinery.objects.filter(owner=owner).delete()
eve_type, _ = EveType.objects.get_or_create_esi(id=35835)
character, _ = EveEntity.objects.get_or_create_esi(id=DUMMY_CHARACTER_ID)
my_extractions = list()
for moon in random.choices(my_moons, k=MAX_REFINERIES):
    if not hasattr(moon, "refinery"):
        print(f"Creating refinery for moon: {moon}")
        refinery = Refinery.objects.create(
            id=moon.eve_moon.id,
            name=f"Test Refinery #{moon.eve_moon.id}",
            moon=moon,
            owner=owner,
            eve_type=eve_type,
        )
        my_extractions.append(
            generate_extraction(
                refinery=refinery,
                chunk_arrival_at=now() + dt.timedelta(days=random.randint(7, 30)),
                started_by=character,
                status=Extraction.Status.STARTED,
            )
        )
        my_extractions.append(
            generate_extraction(
                refinery=refinery,
                chunk_arrival_at=now() - dt.timedelta(days=random.randint(7, 30)),
                started_by=character,
                status=Extraction.Status.CANCELED,
            )
        )
        my_extractions.append(
            generate_extraction(
                refinery=refinery,
                chunk_arrival_at=now() - dt.timedelta(days=random.randint(7, 30)),
                started_by=character,
                status=Extraction.Status.COMPLETED,
            )
        )
print(f"Generating mining ledger...")

for refinery in Refinery.objects.select_related("owner__corporation").filter(
    owner__corporation__corporation_id=DUMMY_CORPORATION_ID
):
    for user in User.objects.order_by("?"):
        for character_ownership in user.character_ownerships.select_related(
            "character"
        ).all()[:MAX_MINING_ALTS_PER_MINER]:
            character, _ = EveEntity.objects.get_or_create_esi(
                id=character_ownership.character.character_id
            )
            corporation, _ = EveEntity.objects.get_or_create_esi(
                id=character_ownership.character.corporation_id
            )
            all_days = sorted([now() - dt.timedelta(days=day) for day in range(31 * 4)])
            for day in random.sample(all_days, k=MAX_MINING_DAYS):
                for ore_type_id in refinery.moon.products.values_list(
                    "ore_type_id", flat=True
                ):
                    MiningLedgerRecord.objects.create(
                        refinery=refinery,
                        day=day,
                        character=character,
                        ore_type_id=ore_type_id,
                        corporation=corporation,
                        quantity=random.randint(10000, 100000),
                        user=user,
                    )


print(f"Updating calculated properties...")
EveMarketPrice.objects.update_from_esi()
for moon in my_moons:
    moon.update_calculated_properties()
for extraction in my_extractions:
    extraction.update_calculated_properties()
print("DONE")



from django.core.management.base import BaseCommand
from eveuniverse.models import EveType
from eveuniverse.tasks import update_or_create_eve_object

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from ... import __title__
from ...constants import EveCategoryId
from ...models import EveOreType
from . import get_input

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Command(BaseCommand):
    help = "Preloads data like ore types from ESI."

    def handle(self, *args, **options):
        self.stdout.write("Loading all ore types from ESI. This can take a while.")
        ore_types_count = EveOreType.objects.count()
        self.stdout.write(
            f"You currently have {ore_types_count} ore types in your database."
        )
        self.stdout.write()
        user_input = get_input("Are you sure you want to proceed? (y/N)?")

        if user_input.lower() == "y":
            self.stdout.write("Tasks for loading ore types have been started.")
            update_or_create_eve_object.delay(
                model_name="EveCategory",
                id=EveCategoryId.ASTEROID.value,
                include_children=True,
                enabled_sections=[
                    EveType.Section.DOGMAS,
                    EveType.Section.TYPE_MATERIALS,
                ],
            )
            self.stdout.write(self.style.SUCCESS("Done"))
        else:
            self.stdout.write(self.style.WARNING("Aborted"))



from enum import IntEnum

# EVE IDs


class EveCategoryId(IntEnum):
    ASTEROID = 25


class EveGroupId(IntEnum):
    MOON = 8
    MINERAL = 18
    REFINERY = 1406
    UBIQUITOUS_MOON_ASTEROIDS = 1884
    COMMON_MOON_ASTEROIDS = 1920
    UNCOMMON_MOON_ASTEROIDS = 1921
    RARE_MOON_ASTEROIDS = 1922
    EXCEPTIONAL_MOON_ASTEROIDS = 1923


class EveTypeId(IntEnum):
    ATHANOR = 35835
    CHROMITE = 45501
    EUXENITE = 45495
    XENOTIME = 45510
    BITUMENS = 45492
    CINNABAR = 45506
    CUBIC_BISTOT = 46676
    FLAWLESS_ARKONOR = 46678
    STABLE_VELDSPAR = 46689
    ZEOLITES = 45490
    BRIMFUL_ZEOLITES = 46280
    GLISTENING_ZEOLITES = 46281

    MOON = 14


class EveDogmaAttributeId(IntEnum):
    ORE_QUALITY = 2699


DATETIME_FORMAT = "%Y-%b-%d %H:%M"
DATE_FORMAT = "%Y-%b-%d"
VALUE_DIVIDER = 1_000_000_000


class IconSize(IntEnum):
    SMALL = 32
    MEDIUM = 64



import datetime as dt

from django.test import TestCase

from ..templatetags import moonmining


class TestFormatisk(TestCase):
    def test_should_return_formatted_string_from_number_1(self):
        # when
        result = moonmining.formatisk(1260000000)
        # then
        self.assertEqual(result, "1.3b")

    def test_should_return_formatted_string_from_number_2(self):
        # when
        result = moonmining.formatisk(123456789)
        # then
        self.assertEqual(result, "123.5m")

    def test_should_return_formatted_string_from_string(self):
        # when
        result = moonmining.formatisk("1234567890")
        # then
        self.assertEqual(result, "1.2b")

    def test_should_return_none_when_type_invalid(self):
        # when
        result = moonmining.formatisk("invalid")
        # then
        self.assertIsNone(result)

    def test_should_use_custom_magnitude(self):
        # when
        result = moonmining.formatisk(123456789, "b")
        # then
        self.assertEqual(result, "0.1b")


class TestDatetime(TestCase):
    def test_should_return_formatted_datetime(self):
        # when
        result = moonmining.datetime(dt.datetime(2021, 3, 24, 14, 50))
        # then
        self.assertEqual(result, "2021-Mar-24 14:50")

    def test_should_return_none_for_invalid_types_1(self):
        # when
        result = moonmining.datetime("invalid")
        # then
        self.assertIsNone(result)

    def test_should_return_none_for_invalid_types_2(self):
        # when
        result = moonmining.datetime(42)
        # then
        self.assertIsNone(result)



# Generated by Django 4.0.3 on 2022-04-21 20:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("moonmining", "0004_add_permission_moon_ledgers"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="eveoretypeextras",
            name="refined_price",
        ),
        migrations.AddField(
            model_name="eveoretypeextras",
            name="current_price",
            field=models.FloatField(
                default=None,
                help_text="price used all price calculations with this type",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="eveoretypeextras",
            name="pricing_method",
            field=models.CharField(
                choices=[
                    ("UN", "Unknown"),
                    ("EC", "Eve client"),
                    ("RP", "Reprocessed materials"),
                ],
                default="UN",
                max_length=2,
            ),
        ),
        migrations.AlterField(
            model_name="moon",
            name="rarity_class",
            field=models.PositiveIntegerField(
                choices=[
                    (0, ""),
                    (4, "R 4"),
                    (8, "R 8"),
                    (16, "R16"),
                    (32, "R32"),
                    (64, "R64"),
                ],
                default=0,
            ),
        ),
    ]



from collections import namedtuple
from typing import Tuple

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction
from django.db.models import ExpressionWrapper, F, FloatField, IntegerField, Sum
from django.db.models.functions import Coalesce
from django.utils.timezone import now
from eveuniverse.managers import EveTypeManager
from eveuniverse.models import EveMoon

from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import (
    MOONMINING_REPROCESSING_YIELD,
    MOONMINING_USE_REPROCESS_PRICING,
)
from .constants import EveCategoryId
from .core import CalculatedExtraction
from .helpers import eveentity_get_or_create_esi_safe

MAX_THREAD_WORKERS = 20
BULK_BATCH_SIZE = 500
logger = LoggerAddTag(get_extension_logger(__name__), __title__)

SurveyProcessResult = namedtuple(
    "SurveyProcessResult", ["moon_name", "success", "error_name"]
)


class EveOreTypeManger(EveTypeManager):
    def get_queryset(self):
        """Return ore types only."""
        return (
            super()
            .get_queryset()
            .select_related("eve_group")
            .filter(published=True)
            .filter(eve_group__eve_category_id=EveCategoryId.ASTEROID)
        )

    def update_current_prices(self, use_process_pricing: bool = None):
        """Update current prices for all ores."""
        from .models import EveOreTypeExtras

        if use_process_pricing is None:
            use_process_pricing = MOONMINING_USE_REPROCESS_PRICING

        for obj in self.filter(published=True).select_related("market_price"):
            if use_process_pricing:
                price = obj.calc_refined_value_per_unit(MOONMINING_REPROCESSING_YIELD)
                pricing_method = EveOreTypeExtras.PricingMethod.REPROCESSED_MATERIALS
            else:
                try:
                    price = obj.market_price.average_price
                    pricing_method = EveOreTypeExtras.PricingMethod.EVE_CLIENT
                except ObjectDoesNotExist:
                    price = None
                    pricing_method = EveOreTypeExtras.PricingMethod.UNKNOWN
            EveOreTypeExtras.objects.update_or_create(
                ore_type=obj,
                defaults={"current_price": price, "pricing_method": pricing_method},
            )


class MiningLedgerRecordManager(models.Manager):
    def get_queryset(self):
        """Add calculated values to all object."""
        sum_price = ExpressionWrapper(
            F("quantity") * Coalesce(F("unit_price"), 0.0),
            output_field=FloatField(),
        )
        sum_volume = ExpressionWrapper(
            F("quantity") * F("ore_type__volume"), output_field=IntegerField()
        )
        return (
            super()
            .get_queryset()
            .select_related("ore_type", "ore_type__extras")
            .annotate(unit_price=F("ore_type__extras__current_price"))
            .annotate(total_price=Sum(sum_price, distinct=True))
            .annotate(total_volume=Sum(sum_volume, distinct=True))
        )


class MoonQuerySet(models.QuerySet):
    def selected_related_defaults(self) -> models.QuerySet:
        return self.select_related(
            "eve_moon",
            "eve_moon__eve_planet__eve_solar_system",
            "eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
            "refinery",
            "refinery__eve_type",
            "refinery__owner",
            "refinery__owner__corporation",
            "refinery__owner__corporation__alliance",
            "label",
        )


class MoonManagerBase(models.Manager):
    def update_moons_from_survey(self, scans: str, user: User = None) -> bool:
        """Update moons from survey input.

        Args:
            scans: raw text input from user containing moon survey data
            user: (optional) user who submitted the data
        """
        surveys, error_name = self._parse_scans(scans)
        if surveys:
            process_results, success = self._process_surveys(surveys, user)
        else:
            process_results = None
            success = False

        if user:
            success = self._send_survey_process_report_to_user(
                process_results, error_name, user, success
            )
        return success

    @staticmethod
    def _parse_scans(scans: str) -> tuple:
        surveys = []
        try:
            lines = scans.split("\n")
            lines_ = []
            for line in lines:
                line = line.strip("\r").split("\t")
                lines_.append(line)
            lines = lines_

            # Find all groups of scans.
            if len(lines[0]) == 0 or lines[0][0] == "Moon":
                lines = lines[1:]
            sublists = []
            for line in lines:
                # Find the lines that start a scan
                if line[0] == "":
                    pass
                else:
                    sublists.append(lines.index(line))

            # Separate out individual surveys
            for i in range(len(sublists)):
                # The First List
                if i == 0:
                    if i + 2 > len(sublists):
                        surveys.append(lines[sublists[i] :])
                    else:
                        surveys.append(lines[sublists[i] : sublists[i + 1]])
                else:
                    if i + 2 > len(sublists):
                        surveys.append(lines[sublists[i] :])
                    else:
                        surveys.append(lines[sublists[i] : sublists[i + 1]])

        except Exception as ex:
            logger.warning(
                "An issue occurred while trying to parse the surveys", exc_info=True
            )
            error_name = type(ex).__name__

        else:
            error_name = ""
        return surveys, error_name

    def _process_surveys(self, surveys: list, user: User) -> Tuple[list, bool]:
        from .models import EveOreType, MoonProduct

        overall_success = True
        process_results = list()
        for survey in surveys:
            moon_name = ""
            try:
                moon_name = survey[0][0]
                moon_id = survey[1][6]
                eve_moon, _ = EveMoon.objects.get_or_create_esi(id=moon_id)
                moon, _ = self.get_or_create(eve_moon=eve_moon)
                moon_products = list()
                survey = survey[1:]
                for product_data in survey:
                    # Trim off the empty index at the front
                    product_data = product_data[1:]
                    ore_type, _ = EveOreType.objects.get_or_create_esi(
                        id=product_data[2]
                    )
                    moon_products.append(
                        MoonProduct(
                            moon=moon, amount=product_data[1], ore_type=ore_type
                        )
                    )
                moon.update_products(moon_products, updated_by=user)
                logger.info("Added moon survey for %s", moon.name)

            except Exception as ex:
                logger.warning(
                    "An issue occurred while processing the following moon survey: %s",
                    survey,
                    exc_info=True,
                )
                error_name = type(ex).__name__
                overall_success = success = False
            else:
                success = True
                error_name = None

            process_results.append(
                SurveyProcessResult(
                    moon_name=moon_name, success=success, error_name=error_name
                )
            )
        return process_results, overall_success

    @staticmethod
    def _send_survey_process_report_to_user(
        process_results: list, error_name: str, user: User, success: bool
    ) -> bool:
        message = "We have completed processing your moon survey input:\n\n"
        if process_results:
            for num, process_result in enumerate(process_results):
                moon_name = process_result.moon_name
                if process_result.success:
                    status = "OK"
                    error_name = ""
                else:
                    status = "FAILED"
                    success = False
                    error_name = "- {}".format(process_result.error_name)
                message += "#{}: {}: {} {}\n".format(
                    num + 1, moon_name, status, error_name
                )
        else:
            message += "\nProcessing failed"

        notify(
            user=user,
            title="Moon survey input processing results: {}".format(
                "OK" if success else "FAILED"
            ),
            message=message,
            level="success" if success else "danger",
        )
        return success


MoonManager = MoonManagerBase.from_queryset(MoonQuerySet)


class ExtractionQuerySet(models.QuerySet):
    def selected_related_defaults(self) -> models.QuerySet:
        return self.select_related(
            "refinery",
            "refinery__moon",
            "refinery__moon__eve_moon",
            "refinery__owner",
            "refinery__owner__corporation",
            "refinery__owner__corporation__alliance",
            "refinery__moon__label",
        )

    def update_status(self):
        """Update status of given extractions according to current time."""
        self.exclude(
            status__in=[self.model.Status.READY, self.model.Status.CANCELED]
        ).filter(
            chunk_arrival_at__lte=now(),
            auto_fracture_at__gt=now(),
        ).update(
            status=self.model.Status.READY
        )
        self.exclude(
            status__in=[self.model.Status.COMPLETED, self.model.Status.CANCELED]
        ).filter(auto_fracture_at__lte=now()).update(status=self.model.Status.COMPLETED)

    def annotate_volume(self) -> models.QuerySet:
        """Add volume of all products"""
        return self.annotate(volume=Sum("products__volume"))


class ExtractionManagerBase(models.Manager):
    def update_from_calculated(self, calculated: CalculatedExtraction) -> bool:
        """Update an extraction object from related calculated extraction
        when there is new information.

        Return True when updated, else False.
        """
        from .models import EveOreType, ExtractionProduct

        if calculated.chunk_arrival_at:
            try:
                extraction = self.get(
                    refinery_id=calculated.refinery_id,
                    chunk_arrival_at=calculated.chunk_arrival_at,
                )
            except self.model.DoesNotExist:
                logger.debug("%s: Could not find matching extraction", calculated)
                return False
        elif calculated.auto_fracture_at:
            try:
                extraction = self.get(
                    refinery_id=calculated.refinery_id,
                    auto_fracture_at=calculated.auto_fracture_at,
                )
            except self.model.DoesNotExist:
                logger.debug("%s: Could not find matching extraction", calculated)
                return False
        else:
            logger.debug(
                "%s: Not enough data to search for matching extraction", calculated
            )
            return False

        needs_update = False
        if calculated.canceled_at and not extraction.canceled_at:
            extraction.canceled_at = calculated.canceled_at
            needs_update = True
        if calculated.canceled_by and not extraction.canceled_by:
            extraction.canceled_by = eveentity_get_or_create_esi_safe(
                calculated.canceled_by
            )
            needs_update = True
        if calculated.canceled_by and not extraction.canceled_by:
            extraction.canceled_by = eveentity_get_or_create_esi_safe(
                calculated.canceled_by
            )
            needs_update = True
        if calculated.fractured_by and not extraction.fractured_by:
            extraction.fractured_by = eveentity_get_or_create_esi_safe(
                calculated.fractured_by
            )
            needs_update = True
        if calculated.fractured_at and not extraction.fractured_at:
            extraction.fractured_at = calculated.fractured_at
            needs_update = True
        if self.model.Status.from_calculated(calculated) != extraction.status:
            extraction.status = self.model.Status.from_calculated(calculated)
            needs_update = True
            status_changed = True
        else:
            status_changed = False
        if calculated.started_by and not extraction.started_by:
            extraction.started_by = eveentity_get_or_create_esi_safe(
                calculated.started_by
            )
            needs_update = True
        updated = False
        if needs_update:
            extraction.save()
            updated = True
        if calculated.products and (status_changed or not extraction.products.exists()):
            # preload eve ore types before transaction starts
            EveOreType.objects.bulk_get_or_create_esi(
                ids=[product.ore_type_id for product in calculated.products]
            )
            products = [
                ExtractionProduct(
                    extraction=extraction,
                    ore_type_id=product.ore_type_id,
                    volume=product.volume,
                )
                for product in calculated.products
            ]
            with transaction.atomic():
                ExtractionProduct.objects.filter(extraction=extraction).delete()
                ExtractionProduct.objects.bulk_create(
                    products, batch_size=BULK_BATCH_SIZE
                )
            extraction.update_calculated_properties()
            updated = True
        return updated


ExtractionManager = ExtractionManagerBase.from_queryset(ExtractionQuerySet)


class RefineryManager(models.Manager):
    def ids(self) -> set:
        return set(self.values_list("id", flat=True))



import json

from eveuniverse.models import EveMoon, EveType
from eveuniverse.tools.testdata import load_testdata_from_dict

from . import test_data_filename


def _load_eveuniverse_from_file():
    with open(test_data_filename(), "r", encoding="utf-8") as f:
        data = json.load(f)

    return data


eveuniverse_testdata = _load_eveuniverse_from_file()


def load_eveuniverse():
    load_testdata_from_dict(eveuniverse_testdata)


def nearest_celestial_stub(eve_solar_system, x, y, z, group_id=None):
    eve_type = EveType.objects.get(id=14)
    if (x, y, z) == (55028384780, 7310316270, -163686684205):
        return eve_solar_system.NearestCelestial(
            eve_type=eve_type,
            eve_object=EveMoon.objects.get(id=40161708),  # Auga V - Moon 1
            distance=123,
        )
    elif (x, y, z) == (45028384780, 6310316270, -163686684205):
        return eve_solar_system.NearestCelestial(
            eve_type=eve_type,
            eve_object=EveMoon.objects.get(id=40161709),  # Auga V - Moon 2
            distance=123,
        )
    return None



from django.contrib import admin
from django.core.exceptions import ObjectDoesNotExist

from . import tasks
from .models import (
    EveOreType,
    EveOreTypeExtras,
    Extraction,
    ExtractionProduct,
    Label,
    MiningLedgerRecord,
    Moon,
    MoonProduct,
    Notification,
    Owner,
    Refinery,
)


class EveOreTypeExtrasInline(admin.StackedInline):
    model = EveOreTypeExtras


@admin.register(EveOreType)
class EveOreTypeAdmin(admin.ModelAdmin):
    list_display = ("name", "_current_price", "_group", "_pricing_method")
    ordering = ("name",)
    list_filter = (
        "extras__pricing_method",
        ("eve_group", admin.RelatedOnlyFieldListFilter),
    )
    search_fields = ("name",)
    inlines = [EveOreTypeExtrasInline]

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("extras", "eve_group")

    @admin.display(ordering="extras__current_price")
    def _current_price(self, obj):
        try:
            return f"{obj.extras.current_price:,.2f}"
        except ObjectDoesNotExist:
            return None

    @admin.display(ordering="eve_group__name")
    def _group(self, obj):
        return str(obj.eve_group)

    @admin.display(ordering="extras__pricing_method")
    def _pricing_method(self, obj):
        try:
            return obj.extras.get_pricing_method_display()
        except ObjectDoesNotExist:
            return None

    def has_add_permission(self, *args, **kwargs) -> bool:
        return False

    def has_change_permission(self, *args, **kwargs) -> bool:
        return False


class ExtractionProductAdmin(admin.TabularInline):
    model = ExtractionProduct


@admin.register(Extraction)
class ExtractionAdmin(admin.ModelAdmin):
    list_display = ("chunk_arrival_at", "status", "_owner", "refinery", "_ledger")
    ordering = ("-chunk_arrival_at",)
    list_filter = ("chunk_arrival_at", "status", "refinery__owner", "refinery")
    search_fields = ("refinery__moon__eve_moon__name",)
    inlines = [ExtractionProductAdmin]
    actions = ["update_calculated_properties"]

    @admin.display(description="Update calculated properties for selected extrations.")
    def update_calculated_properties(self, request, queryset):
        num = 0
        for obj in queryset:
            tasks.update_extraction_calculated_properties.delay(extraction_pk=obj.pk)
            num += 1
        self.message_user(
            request, f"Started updating calculated properties for {num} extractions."
        )

    def _owner(self, obj):
        return obj.refinery.owner

    @admin.display(boolean=True)
    def _ledger(self, obj):
        if obj.status != Extraction.Status.COMPLETED:
            return None
        return obj.ledger.exists()

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False


@admin.register(Label)
class LabelAdmin(admin.ModelAdmin):
    list_display = ("name", "style")
    fields = ("name", "description", "style")


@admin.register(MiningLedgerRecord)
class MiningLedgerRecordAdmin(admin.ModelAdmin):
    list_display = ("refinery", "day", "user", "character", "ore_type", "quantity")
    ordering = ["refinery", "day", "user", "character", "ore_type"]
    list_filter = (
        "refinery",
        "day",
        "user",
        ("character", admin.RelatedOnlyFieldListFilter),
        "ore_type",
    )

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("refinery", "character", "ore_type", "user")

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


class MoonHasRefineryFilter(admin.SimpleListFilter):
    title = "has refinery"
    parameter_name = "has_refinery"

    def lookups(self, request, model_admin):
        return (
            ("yes", "Yes"),
            ("no", "No"),
        )

    def queryset(self, request, queryset):
        """Return the filtered queryset"""
        if self.value() == "yes":
            return queryset.filter(refinery__isnull=False)
        elif self.value() == "no":
            return queryset.filter(refinery__isnull=True)
        return queryset


class MoonProductAdminInline(admin.TabularInline):
    model = MoonProduct

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(Moon)
class MoonAdmin(admin.ModelAdmin):
    list_display = (
        "eve_moon",
        "_constellation",
        "_region",
        "label",
        "products_updated_at",
        "_refinery",
        "_owner",
    )
    list_filter = (
        "rarity_class",
        MoonHasRefineryFilter,
        "label",
        "products_updated_at",
        ("refinery__owner", admin.RelatedOnlyFieldListFilter),
        (
            "eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
            admin.RelatedOnlyFieldListFilter,
        ),
        (
            "eve_moon__eve_planet__eve_solar_system__eve_constellation",
            admin.RelatedOnlyFieldListFilter,
        ),
        ("eve_moon__eve_planet__eve_solar_system", admin.RelatedOnlyFieldListFilter),
    )
    actions = ["update_calculated_properties", "update_products_from_latest_extraction"]
    inlines = (MoonProductAdminInline,)
    fields = (
        "eve_moon",
        "rarity_class",
        "value",
        "label",
        "products_updated_at",
        "products_updated_by",
    )
    readonly_fields = (
        "eve_moon",
        "products_updated_at",
        "products_updated_by",
        "rarity_class",
        "value",
    )
    search_fields = [
        "eve_moon__name",
        "refinery__name",
        "refinery__owner__corporation__corporation_name",
        "refinery__owner__corporation__alliance__alliance_name",
    ]

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related(
            "refinery",
            "refinery__owner",
            "refinery__owner__corporation",
            "refinery__owner__corporation__alliance",
            "eve_moon",
            "eve_moon__eve_planet__eve_solar_system",
            "eve_moon__eve_planet__eve_solar_system__eve_constellation",
            "eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region",
        )

    def has_add_permission(self, request):
        return False

    @admin.display(ordering="eve_moon__eve_planet__eve_solar_system")
    def _solar_system(self, obj) -> str:
        return obj.eve_moon.eve_planet.eve_solar_system

    @admin.display(ordering="eve_moon__eve_planet__eve_solar_system__eve_constellation")
    def _constellation(self, obj) -> str:
        return obj.eve_moon.eve_planet.eve_solar_system.eve_constellation

    @admin.display(
        ordering="eve_moon__eve_planet__eve_solar_system__eve_constellation__eve_region"
    )
    def _region(self, obj) -> str:
        return obj.eve_moon.eve_planet.eve_solar_system.eve_constellation.eve_region

    @admin.display(ordering="refinery__name")
    def _refinery(self, obj) -> str:
        return obj.refinery.name

    @admin.display(ordering="refinery__owner__name")
    def _owner(self, obj) -> str:
        return obj.refinery.owner.name

    @admin.display(description="Update calculated properties for selected moons.")
    def update_calculated_properties(self, request, queryset):
        num = 0
        for obj in queryset:
            tasks.update_moon_calculated_properties.delay(moon_pk=obj.pk)
            num += 1
        self.message_user(
            request, f"Started updating calculated properties for {num} moons."
        )

    @admin.display(
        description=("Update products from latest extraction for selected moons.")
    )
    def update_products_from_latest_extraction(self, request, queryset):
        num = 0
        for obj in queryset:
            tasks.update_moon_products_from_latest_extraction.delay(moon_pk=obj.pk)
            num += 1
        self.message_user(
            request,
            f"Started updating products from latest extractions for {num} moons.",
        )


@admin.register(Notification)
class NotificationAdmin(admin.ModelAdmin):
    list_display = (
        "notification_id",
        "owner",
        "notif_type",
        "timestamp",
        "created",
        "last_updated",
    )
    ordering = ["-timestamp"]
    list_filter = ("owner", "notif_type")

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False


@admin.register(Owner)
class OwnerAdmin(admin.ModelAdmin):
    list_display = (
        "__str__",
        "_alliance",
        "character_ownership",
        "is_enabled",
        "last_update_at",
        "last_update_ok",
    )
    ordering = ["corporation"]
    search_fields = ("refinery__moon__eve_moon__name",)
    list_filter = (
        "is_enabled",
        "last_update_ok",
        "corporation__alliance",
    )
    actions = ["update_owner"]

    @admin.display(ordering="corporation__alliance__alliance_name")
    def _alliance(self, obj):
        return obj.corporation.alliance

    @admin.display(description="Update selected owners from ESI")
    def update_owner(self, request, queryset):
        for obj in queryset:
            tasks.update_owner.delay(obj.pk)
            text = f"Started updating owner: {obj}. "
            self.message_user(request, text)

    def has_add_permission(self, request):
        return False

    def get_readonly_fields(self, request, obj=None):
        if obj:  # editing an existing object
            return self.readonly_fields + (
                "corporation",
                "character_ownership",
                "last_update_at",
                "last_update_ok",
            )
        return self.readonly_fields


@admin.register(Refinery)
class RefineryAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "moon",
        "owner",
        "eve_type",
        "ledger_last_update_ok",
        "ledger_last_update_at",
    )
    ordering = ["name"]
    list_filter = (
        ("owner__corporation", admin.RelatedOnlyFieldListFilter),
        "ledger_last_update_ok",
    )

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related(
            "moon", "moon__eve_moon", "eve_type", "owner", "owner__corporation"
        )

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False

    actions = ["update_mining_ledger"]

    @admin.display(description="Update mining ledger for selected refineries from ESI")
    def update_mining_ledger(self, request, queryset):
        for obj in queryset:
            tasks.update_mining_ledger_for_refinery.delay(obj.id)
            text = f"Started updating mining ledger: {obj}. "
            self.message_user(request, text)



import json

from django.http import JsonResponse
from eveuniverse.models import EveEntity, EveMarketPrice, EveType

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter, response_text

from ..constants import EveTypeId
from ..models import EveOreType, Owner


def create_default_user_from_evecharacter(character_id):
    return create_user_from_evecharacter(
        character_id,
        permissions=[
            "moonmining.basic_access",
            "moonmining.upload_moon_scan",
            "moonmining.extractions_access",
            "moonmining.add_refinery_owner",
        ],
        scopes=Owner.esi_scopes(),
    )


def generate_eve_entities_from_allianceauth():
    for character in EveCharacter.objects.all():
        EveEntity.objects.create(
            id=character.character_id,
            name=character.character_name,
            category=EveEntity.CATEGORY_CHARACTER,
        )
        EveEntity.objects.get_or_create(
            id=character.corporation_id,
            name=character.corporation_name,
            category=EveEntity.CATEGORY_CORPORATION,
        )
        if character.alliance_id:
            EveEntity.objects.get_or_create(
                id=character.alliance_id,
                name=character.alliance_name,
                category=EveEntity.CATEGORY_ALLIANCE,
            )


def generate_market_prices(use_process_pricing=False):
    tungsten = EveType.objects.get(id=16637)
    EveMarketPrice.objects.create(eve_type=tungsten, average_price=7000)
    mercury = EveType.objects.get(id=16646)
    EveMarketPrice.objects.create(eve_type=mercury, average_price=9750)
    evaporite_deposits = EveType.objects.get(id=16635)
    EveMarketPrice.objects.create(eve_type=evaporite_deposits, average_price=950)
    pyerite = EveType.objects.get(id=35)
    EveMarketPrice.objects.create(eve_type=pyerite, average_price=10)
    zydrine = EveType.objects.get(id=39)
    EveMarketPrice.objects.create(eve_type=zydrine, average_price=1.7)
    megacyte = EveType.objects.get(id=40)
    EveMarketPrice.objects.create(eve_type=megacyte, average_price=640)
    tritanium = EveType.objects.get(id=34)
    EveMarketPrice.objects.create(eve_type=tritanium, average_price=5)
    mexallon = EveType.objects.get(id=36)
    EveMarketPrice.objects.create(eve_type=mexallon, average_price=117.0)
    EveMarketPrice.objects.create(eve_type_id=45506, average_price=2400.0)
    EveMarketPrice.objects.create(eve_type_id=46676, average_price=609.0)
    EveMarketPrice.objects.create(eve_type_id=46678, average_price=310.9)
    EveMarketPrice.objects.create(eve_type_id=46689, average_price=7.7)

    EveMarketPrice.objects.create(eve_type_id=EveTypeId.CHROMITE, average_price=2_000)
    EveMarketPrice.objects.create(eve_type_id=EveTypeId.EUXENITE, average_price=1_300)
    EveMarketPrice.objects.create(eve_type_id=EveTypeId.XENOTIME, average_price=10_000)
    EveOreType.objects.update_current_prices(use_process_pricing=use_process_pricing)


def json_response_to_python_2(response: JsonResponse, data_key="data") -> object:
    """Convert JSON response into Python object."""
    data = json.loads(response_text(response))
    return data[data_key]


def json_response_to_dict_2(response: JsonResponse, key="id", data_key="data") -> dict:
    """Convert JSON response into dict by given key."""
    return {x[key]: x for x in json_response_to_python_2(response, data_key)}



import logging

from django.core.management.base import BaseCommand

from app_utils.logging import LoggerAddTag

from moonmining.models import Extraction, Moon

from ... import __title__, tasks
from . import get_input

logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class Command(BaseCommand):
    help = "Calculate all properties for moons and extractions."

    def handle(self, *args, **options):
        moon_count = Moon.objects.count()
        extractions_count = Extraction.objects.count()
        self.stdout.write(
            f"Updating calculated properties for {moon_count} moons "
            f"and {extractions_count} extractions. This can take a while."
        )
        self.stdout.write()
        user_input = get_input("Are you sure you want to proceed? (Y/n)?")

        if user_input.lower() != "n":
            tasks.run_calculated_properties_update.delay()
            self.stdout.write(self.style.SUCCESS("Update started."))
        else:
            self.stdout.write(self.style.WARNING("Aborted"))



import datetime as dt
import random
from typing import List

import factory
import factory.fuzzy
import pytz

from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveMoon, EveType

from allianceauth.eveonline.models import EveCorporationInfo
from app_utils.testdata_factories import (
    EveCharacterFactory,
    EveCorporationInfoFactory,
    UserMainFactory,
)
from app_utils.testing import create_user_from_evecharacter

from ...app_settings import MOONMINING_VOLUME_PER_DAY
from ...constants import EveTypeId
from ...core import CalculatedExtraction, CalculatedExtractionProduct
from ...models import (
    EveOreType,
    Extraction,
    ExtractionProduct,
    MiningLedgerRecord,
    Moon,
    MoonProduct,
    Notification,
    NotificationType,
    Owner,
    Refinery,
)

FUZZY_START_YEAR = 2008


def datetime_to_ldap(my_dt: dt.datetime) -> int:
    """datetime.datetime to ldap"""
    return (
        ((my_dt - dt.datetime(1970, 1, 1, tzinfo=pytz.utc)).total_seconds())
        + 11644473600
    ) * 10000000


# Auth


class DefaultOwnerUserMainFactory(UserMainFactory):
    main_character__scopes = Owner.esi_scopes()
    permissions__ = [
        "moonmining.basic_access",
        "moonmining.upload_moon_scan",
        "moonmining.extractions_access",
        "moonmining.add_refinery_owner",
    ]

    @factory.lazy_attribute
    def main_character__character(self):
        corporation = EveCorporationInfoFactory(
            corporation_id=2001, corporation_name="Wayne Technologies"
        )
        return EveCharacterFactory(
            character_id=1001, character_name="Bruce Wayne", corporation=corporation
        )


# eveuniverse


class EveEntityFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = EveEntity
        django_get_or_create = ("id", "name")

    id = factory.Sequence(lambda n: 10_001 + n)


class EveEntityCharacterFactory(EveEntityFactory):
    name = factory.Faker("name")
    category = EveEntity.CATEGORY_CHARACTER


class EveEntityCorporationFactory(EveEntityFactory):
    name = factory.Faker("company")
    category = EveEntity.CATEGORY_CORPORATION


class EveEntityAllianceFactory(EveEntityFactory):
    name = factory.Faker("company")
    category = EveEntity.CATEGORY_ALLIANCE


# moonmining


def random_percentages(num_parts: int) -> List[float]:
    percentages = []
    total = 0
    for _ in range(num_parts - 1):
        part = random.randint(0, 100 - total)
        percentages.append(part)
        total += part
    percentages.append((100 - total) / 100)
    return percentages


def _generate_calculated_extraction_products(
    extraction: CalculatedExtraction,
) -> List[CalculatedExtractionProduct]:
    ore_type_ids = [EveTypeId.CHROMITE, EveTypeId.EUXENITE, EveTypeId.XENOTIME]
    percentages = random_percentages(3)
    duration = (
        (extraction.chunk_arrival_at - extraction.started_at).total_seconds()
        / 3600
        / 24
    )
    products = [
        CalculatedExtractionProductFactory(
            ore_type_id=ore_type_id,
            volume=percentages.pop() * MOONMINING_VOLUME_PER_DAY * duration,
        )
        for ore_type_id in ore_type_ids
    ]
    return products


class CalculatedExtractionProductFactory(factory.Factory):
    class Meta:
        model = CalculatedExtractionProduct


class CalculatedExtractionFactory(factory.Factory):
    class Meta:
        model = CalculatedExtraction

    auto_fracture_at = factory.LazyAttribute(
        lambda obj: obj.chunk_arrival_at + dt.timedelta(hours=3)
    )
    chunk_arrival_at = factory.LazyAttribute(
        lambda obj: obj.started_at + dt.timedelta(days=20)
    )
    refinery_id = factory.Sequence(lambda n: n + 1800000000001)
    status = CalculatedExtraction.Status.STARTED
    started_at = factory.fuzzy.FuzzyDateTime(
        dt.datetime(FUZZY_START_YEAR, 1, 1, tzinfo=pytz.utc), force_microsecond=0
    )

    @factory.lazy_attribute
    def started_by(self):
        character = EveEntityCharacterFactory(name="Bruce Wayne")
        return character.id

    @factory.lazy_attribute
    def products(self):
        return _generate_calculated_extraction_products(self)


class MiningLedgerRecordFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MiningLedgerRecord

    day = factory.fuzzy.FuzzyDate((now() - dt.timedelta(days=120)).date())
    character = factory.SubFactory(EveEntityCharacterFactory)
    corporation = factory.SubFactory(EveEntityCorporationFactory)
    ore_type = factory.LazyFunction(lambda: EveOreType.objects.order_by("?").first())
    quantity = factory.fuzzy.FuzzyInteger(10000)


class MoonFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Moon
        exclude = ("create_products",)

    products_updated_at = factory.fuzzy.FuzzyDateTime(
        dt.datetime(FUZZY_START_YEAR, 1, 1, tzinfo=pytz.utc), force_microsecond=0
    )

    @factory.lazy_attribute
    def eve_moon(self):
        return EveMoon.objects.exclude(
            id__in=list(Moon.objects.values_list("eve_moon_id", flat=True))
        ).first()

    @factory.post_generation
    def create_products(obj, create, extracted, **kwargs):
        """Set this param to False to disable."""
        if not create or extracted is False:
            return
        ore_type_ids = [EveTypeId.CHROMITE, EveTypeId.EUXENITE, EveTypeId.XENOTIME]
        percentages = random_percentages(3)
        for ore_type_id in ore_type_ids:
            ore_type, _ = EveOreType.objects.get_or_create_esi(id=ore_type_id)
            MoonProductFactory(moon=obj, ore_type=ore_type, amount=percentages.pop())
        obj.update_calculated_properties()


class MoonProductFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MoonProduct


class OwnerFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Owner

    last_update_at = factory.LazyFunction(now)
    last_update_ok = True

    @factory.lazy_attribute
    def character_ownership(self):
        _, obj = create_user_from_evecharacter(
            1001,
            permissions=[
                "moonmining.basic_access",
                "moonmining.upload_moon_scan",
                "moonmining.extractions_access",
                "moonmining.add_refinery_owner",
            ],
            scopes=Owner.esi_scopes(),
        )
        return obj

    @factory.lazy_attribute
    def corporation(self):
        corporation_id = (
            self.character_ownership.character.corporation_id
            if self.character_ownership
            else 2001
        )
        return EveCorporationInfo.objects.get(corporation_id=corporation_id)


class RefineryFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Refinery

    id = factory.Sequence(lambda n: n + 1900000000001)
    name = factory.Faker("city")
    moon = factory.SubFactory(MoonFactory)
    owner = factory.SubFactory(OwnerFactory)

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(id=EveTypeId.ATHANOR)


class ExtractionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Extraction

    started_at = factory.fuzzy.FuzzyDateTime(
        dt.datetime(FUZZY_START_YEAR, 1, 1, tzinfo=pytz.utc), force_microsecond=0
    )
    chunk_arrival_at = factory.LazyAttribute(
        lambda obj: obj.started_at + dt.timedelta(days=20)
    )
    auto_fracture_at = factory.LazyAttribute(
        lambda obj: obj.chunk_arrival_at + dt.timedelta(hours=3)
    )
    refinery = factory.SubFactory(RefineryFactory)
    status = Extraction.Status.STARTED

    @factory.post_generation
    def create_products(obj, create, extracted, **kwargs):
        """Set this param to False to disable."""
        if not create or extracted is False:
            return
        for product in obj.refinery.moon.products.all():
            ExtractionProductFactory(
                extraction=obj,
                ore_type=product.ore_type,
                volume=MOONMINING_VOLUME_PER_DAY
                * obj.duration_in_days
                * product.amount,
            )
        obj.update_calculated_properties()


class ExtractionProductFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ExtractionProduct


class NotificationFactory(factory.django.DjangoModelFactory):
    """Create notifications from Extraction objects."""

    class Meta:
        model = Notification

    class Params:
        extraction = factory.SubFactory(ExtractionFactory)

    notification_id = factory.Sequence(lambda n: 1_900_000_001 + n)
    owner = factory.LazyAttribute(lambda obj: obj.extraction.refinery.owner)
    created = factory.fuzzy.FuzzyDateTime(
        dt.datetime(FUZZY_START_YEAR, 1, 1, tzinfo=pytz.utc), force_microsecond=0
    )
    notif_type = factory.LazyAttribute(
        lambda obj: obj.extraction.status_enum.to_notification_type
    )
    last_updated = factory.LazyFunction(now)
    sender = factory.SubFactory(EveEntityCorporationFactory, name="DED")
    timestamp = factory.LazyAttribute(lambda obj: obj.extraction.started_at)

    @factory.lazy_attribute
    def details(self):
        def _details_link(character: EveEntity) -> str:
            return f'<a href="showinfo:1379//{character.id}">{character.name}</a>'

        def _to_ore_volume_by_type(extraction):
            return {
                str(obj.ore_type_id): obj.volume for obj in extraction.products.all()
            }

        refinery = self.extraction.refinery
        data = {
            "moonID": self.extraction.refinery.moon.eve_moon_id,
            "structureID": self.extraction.refinery_id,
            "solarSystemID": refinery.moon.solar_system().id,
            "structureLink": (
                f'<a href="showinfo:35835//{refinery.id}">{refinery.name}</a>'
            ),
            "structureName": refinery.name,
            "structureTypeID": refinery.eve_type_id,
        }
        if self.extraction.status == Extraction.Status.STARTED:
            started_by = (
                self.extraction.started_by
                if self.extraction.started_by
                else EveEntityCharacterFactory()
            )
            data.update(
                {
                    "autoTime": datetime_to_ldap(self.extraction.auto_fracture_at),
                    "readyTime": datetime_to_ldap(self.extraction.chunk_arrival_at),
                    "startedBy": started_by.id,
                    "startedByLink": _details_link(started_by),
                    "oreVolumeByType": _to_ore_volume_by_type(self.extraction),
                }
            )
        elif self.extraction.status == Extraction.Status.READY:
            data.update(
                {
                    "autoTime": datetime_to_ldap(self.extraction.auto_fracture_at),
                    "oreVolumeByType": _to_ore_volume_by_type(self.extraction),
                }
            )
        elif self.extraction.status == Extraction.Status.COMPLETED:
            data.update(
                {
                    "oreVolumeByType": _to_ore_volume_by_type(self.extraction),
                }
            )

        elif self.extraction.status == Extraction.Status.COMPLETED:
            fired_by = EveEntityCharacterFactory()
            data.update(
                {
                    "firedBy": fired_by.id,
                    "firedByLink": _details_link(fired_by),
                    "oreVolumeByType": _to_ore_volume_by_type(self.extraction),
                }
            )
        elif self.extraction.status == Extraction.Status.CANCELED:
            canceled_by = (
                self.extraction.canceled_by
                if self.extraction.canceled_by
                else EveEntityCharacterFactory()
            )
            data.update(
                {
                    "cancelledBy": canceled_by.id,
                    "cancelledByLink": _details_link(canceled_by),
                }
            )
        return data


class NotificationFactory2(factory.django.DjangoModelFactory):
    """Create notifications from CalculatedExtraction objects."""

    class Meta:
        model = Notification
        exclude = (
            "extraction",
            "moon_id",
            "solar_system_id",
            "structure_id",
            "structure_name",
            "structure_type_id",
        )

    class Params:
        create_products = False

    # regular
    notification_id = factory.Sequence(lambda n: 1_900_000_001 + n)
    owner = factory.SubFactory(OwnerFactory)
    created = factory.fuzzy.FuzzyDateTime(
        dt.datetime(FUZZY_START_YEAR, 1, 1, tzinfo=pytz.utc), force_microsecond=0
    )
    last_updated = factory.LazyFunction(now)
    sender = factory.SubFactory(EveEntityCorporationFactory, name="DED")
    timestamp = factory.LazyAttribute(lambda obj: obj.extraction.started_at)

    # excluded
    extraction = factory.SubFactory(CalculatedExtractionFactory)
    moon_id = 40161708  # Auga V - Moon 1
    solar_system_id = 30002542  # Auga V
    structure_name = factory.Faker("city")
    structure_type_id = EveTypeId.ATHANOR

    @factory.lazy_attribute
    def notif_type(self):
        status_map = {
            CalculatedExtraction.Status.STARTED: (
                NotificationType.MOONMINING_EXTRACTION_STARTED
            ),
            CalculatedExtraction.Status.READY: (
                NotificationType.MOONMINING_EXTRACTION_FINISHED
            ),
            CalculatedExtraction.Status.COMPLETED: (
                NotificationType.MOONMINING_LASER_FIRED
            ),
            CalculatedExtraction.Status.CANCELED: (
                NotificationType.MOONMINING_EXTRACTION_CANCELLED
            ),
        }
        try:
            return status_map[self.extraction.status]
        except KeyError:
            raise ValueError(f"Invalid status: {self.extraction.status}") from None

    @factory.lazy_attribute
    def details(self):
        def _details_link(character: EveEntity) -> str:
            return f'<a href="showinfo:1379//{character.id}">{character.name}</a>'

        def _to_ore_volume_by_type(extraction):
            return {str(obj.ore_type_id): obj.volume for obj in extraction.products}

        if self.create_products:
            self.extraction.products = _generate_calculated_extraction_products(
                self.extraction
            )

        data = {
            "moonID": self.moon_id,
            "structureID": self.extraction.refinery_id,
            "solarSystemID": self.solar_system_id,
            "structureLink": (
                f'<a href="showinfo:35835//{self.extraction.refinery_id}">{self.structure_name}</a>'
            ),
            "structureName": self.structure_name,
            "structureTypeID": self.structure_type_id,
        }
        if self.extraction.status == CalculatedExtraction.Status.STARTED:
            started_by = (
                EveEntityCharacterFactory(id=self.extraction.started_by)
                if self.extraction.started_by
                else EveEntityCharacterFactory()
            )
            data.update(
                {
                    "autoTime": datetime_to_ldap(self.extraction.auto_fracture_at),
                    "readyTime": datetime_to_ldap(self.extraction.chunk_arrival_at),
                    "startedBy": started_by.id,
                    "startedByLink": _details_link(started_by),
                    "oreVolumeByType": _to_ore_volume_by_type(self.extraction),
                }
            )
        elif self.extraction.status == CalculatedExtraction.Status.READY:
            data.update(
                {
                    "autoTime": datetime_to_ldap(self.extraction.auto_fracture_at),
                    "oreVolumeByType": _to_ore_volume_by_type(self.extraction),
                }
            )
        elif self.extraction.status == CalculatedExtraction.Status.COMPLETED:
            data.update(
                {
                    "oreVolumeByType": _to_ore_volume_by_type(self.extraction),
                }
            )

        elif self.extraction.status == CalculatedExtraction.Status.COMPLETED:
            fired_by = (
                EveEntityCharacterFactory(id=self.extraction.fractured_by)
                if self.extraction.fractured_by
                else EveEntityCharacterFactory()
            )
            data.update(
                {
                    "firedBy": fired_by.id,
                    "firedByLink": _details_link(fired_by),
                    "oreVolumeByType": _to_ore_volume_by_type(self.extraction),
                }
            )
        elif self.extraction.status == CalculatedExtraction.Status.CANCELED:
            canceled_by = (
                EveEntityCharacterFactory(id=self.extraction.canceled_by)
                if self.extraction.canceled_by
                else EveEntityCharacterFactory()
            )
            data.update(
                {
                    "cancelledBy": canceled_by.id,
                    "cancelledByLink": _details_link(canceled_by),
                }
            )
        return data



from io import StringIO
from pathlib import Path
from unittest.mock import patch

from django.core.management import CommandError, call_command
from django.test import override_settings
from eveuniverse.models import EveMarketPrice, EveType

from app_utils.testing import NoSocketsTestCase

from ..models import Moon
from .testdata.esi_client_stub import esi_client_stub
from .testdata.load_eveuniverse import load_eveuniverse

MODELS_PATH = "moonmining.models"
PACKAGE_PATH = "moonmining.management.commands"


@patch(MODELS_PATH + ".esi")
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestImportMoons(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.import_file = Path(__file__).parent / "testdata" / "moons_for_import.csv"

    def setUp(self) -> None:
        self.out = StringIO()

    @patch(PACKAGE_PATH + ".moonmining_import_moons.is_esi_online", new=lambda: True)
    def test_should_create_moons(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        tungsten = EveType.objects.get(id=16637)
        mercury = EveType.objects.get(id=16646)
        evaporite_deposits = EveType.objects.get(id=16635)
        EveMarketPrice.objects.create(eve_type=tungsten, average_price=7000)
        EveMarketPrice.objects.create(eve_type=mercury, average_price=9750)
        EveMarketPrice.objects.create(eve_type=evaporite_deposits, average_price=950)
        # when
        call_command("moonmining_import_moons", str(self.import_file), stdout=self.out)
        # then
        m1 = Moon.objects.get(pk=40161708)
        self.assertIsNone(m1.products_updated_by)
        self.assertIsNone(m1.products_updated_at)
        self.assertEqual(m1.products.count(), 4)
        self.assertEqual(m1.products.get(ore_type_id=45506).amount, 0.19)
        self.assertEqual(m1.products.get(ore_type_id=46676).amount, 0.23)
        self.assertEqual(m1.products.get(ore_type_id=46678).amount, 0.25)
        self.assertEqual(m1.products.get(ore_type_id=46689).amount, 0.33)

        m2 = Moon.objects.get(pk=40161709)
        self.assertIsNone(m2.products_updated_by)
        self.assertIsNone(m2.products_updated_at)
        self.assertEqual(m2.products.count(), 4)
        self.assertEqual(m2.products.get(ore_type_id=45492).amount, 0.27)
        self.assertEqual(m2.products.get(ore_type_id=45494).amount, 0.23)
        self.assertEqual(m2.products.get(ore_type_id=46676).amount, 0.21)
        self.assertEqual(m2.products.get(ore_type_id=46678).amount, 0.29)

    @patch(PACKAGE_PATH + ".moonmining_import_moons.is_esi_online", new=lambda: True)
    def test_should_abort_when_input_file_not_found(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        import_file = Path(__file__).parent / "testdata" / "unknown_file.xyz"
        # when/then
        with self.assertRaises(CommandError):
            call_command("moonmining_import_moons", str(import_file), stdout=self.out)

    @patch(PACKAGE_PATH + ".moonmining_import_moons.is_esi_online", new=lambda: False)
    def test_should_abort_when_esi_is_offline(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when/then
        with self.assertRaises(CommandError):
            call_command(
                "moonmining_import_moons", str(self.import_file), stdout=self.out
            )



# Generated by Django 3.1.6 on 2021-04-06 21:45

import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("authentication", "0017_remove_fleetup_permission"),
        ("eveonline", "0014_auto_20210105_1413"),
        ("eveuniverse", "0005_type_materials_and_sections"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access the moonmining app"),
                    (
                        "extractions_access",
                        "Can access extractions and view owned moons",
                    ),
                    ("reports_access", "Can access reports"),
                    ("view_all_moons", "Can view all known moons"),
                    ("upload_moon_scan", "Can upload moon scans"),
                    ("add_refinery_owner", "Can add refinery owner"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Extraction",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "started_at",
                    models.DateTimeField(help_text="when this extraction was started"),
                ),
                (
                    "auto_fracture_at",
                    models.DateTimeField(
                        help_text="when this extraction will be automatically fractured"
                    ),
                ),
                (
                    "canceled_at",
                    models.DateTimeField(
                        default=None,
                        help_text="when this extraction was canceled",
                        null=True,
                    ),
                ),
                (
                    "chunk_arrival_at",
                    models.DateTimeField(
                        db_index=True,
                        help_text="when this extraction is ready to be fractured",
                    ),
                ),
                (
                    "fractured_at",
                    models.DateTimeField(
                        default=None,
                        help_text="when this extraction was fractured",
                        null=True,
                    ),
                ),
                (
                    "is_jackpot",
                    models.BooleanField(
                        default=None,
                        help_text="Whether this is a jackpot extraction (calculated)",
                        null=True,
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("ST", "started"),
                            ("CN", "canceled"),
                            ("RD", "ready"),
                            ("CP", "completed"),
                            ("UN", "undefined"),
                        ],
                        db_index=True,
                        default="UN",
                        max_length=2,
                    ),
                ),
                (
                    "value",
                    models.FloatField(
                        default=None,
                        help_text="Estimated value of this extraction (calculated)",
                        null=True,
                        validators=[django.core.validators.MinValueValidator(0.0)],
                    ),
                ),
                (
                    "canceled_by",
                    models.ForeignKey(
                        default=None,
                        help_text="Eve character who canceled this extraction",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "fractured_by",
                    models.ForeignKey(
                        default=None,
                        help_text="Eve character who fractured this extraction (if any)",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Moon",
            fields=[
                (
                    "eve_moon",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="known_moon",
                        serialize=False,
                        to="eveuniverse.evemoon",
                    ),
                ),
                (
                    "products_updated_at",
                    models.DateTimeField(
                        default=None,
                        help_text="Time the last moon survey was uploaded",
                        null=True,
                    ),
                ),
                (
                    "rarity_class",
                    models.PositiveIntegerField(
                        choices=[
                            (0, ""),
                            (4, "R4"),
                            (8, "R8"),
                            (16, "R16"),
                            (32, "R32"),
                            (64, "R64"),
                        ],
                        default=0,
                    ),
                ),
                (
                    "value",
                    models.FloatField(
                        db_index=True,
                        default=None,
                        help_text="Calculated value estimate",
                        null=True,
                        validators=[django.core.validators.MinValueValidator(0.0)],
                    ),
                ),
                (
                    "products_updated_by",
                    models.ForeignKey(
                        default=None,
                        help_text="User who uploaded the last moon survey",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Owner",
            fields=[
                (
                    "corporation",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="mining_corporation",
                        serialize=False,
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        db_index=True,
                        default=True,
                        help_text="disabled corporations are excluded from the update process",
                    ),
                ),
                (
                    "last_update_at",
                    models.DateTimeField(
                        default=None,
                        help_text="time of last successful update",
                        null=True,
                    ),
                ),
                (
                    "last_update_ok",
                    models.BooleanField(
                        default=None,
                        help_text="True if the last update was successful",
                        null=True,
                    ),
                ),
                (
                    "character_ownership",
                    models.ForeignKey(
                        default=None,
                        help_text="character used to sync this corporation from ESI",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="authentication.characterownership",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EveOreType",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("eveuniverse.evetype",),
        ),
        migrations.CreateModel(
            name="Refinery",
            fields=[
                (
                    "id",
                    models.PositiveBigIntegerField(primary_key=True, serialize=False),
                ),
                ("name", models.CharField(db_index=True, max_length=150)),
                (
                    "eve_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "moon",
                    models.OneToOneField(
                        default=None,
                        help_text="The moon this refinery is anchored at (if any)",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="refinery",
                        to="moonmining.moon",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        help_text="Corporation that owns this refinery",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="refineries",
                        to="moonmining.owner",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("notification_id", models.PositiveBigIntegerField(verbose_name="id")),
                (
                    "created",
                    models.DateTimeField(
                        default=None,
                        help_text="Date when this notification was first received from ESI",
                        null=True,
                    ),
                ),
                ("details", models.JSONField(default=dict)),
                (
                    "notif_type",
                    models.CharField(
                        db_index=True,
                        default="",
                        help_text="type of this notification as reported by ESI",
                        max_length=100,
                        verbose_name="type",
                    ),
                ),
                (
                    "is_read",
                    models.BooleanField(
                        default=None,
                        help_text="True when this notification has read in the eve client",
                        null=True,
                    ),
                ),
                (
                    "last_updated",
                    models.DateTimeField(
                        help_text="Date when this notification has last been updated from ESI"
                    ),
                ),
                ("timestamp", models.DateTimeField(db_index=True)),
                (
                    "owner",
                    models.ForeignKey(
                        help_text="Corporation that received this notification",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="notifications",
                        to="moonmining.owner",
                    ),
                ),
                (
                    "sender",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="MoonProduct",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "amount",
                    models.FloatField(
                        validators=[
                            django.core.validators.MinValueValidator(0.0),
                            django.core.validators.MaxValueValidator(1.0),
                        ]
                    ),
                ),
                (
                    "moon",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="products",
                        to="moonmining.moon",
                    ),
                ),
                (
                    "ore_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="moonmining.eveoretype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ExtractionProduct",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "volume",
                    models.FloatField(
                        validators=[django.core.validators.MinValueValidator(0.0)]
                    ),
                ),
                (
                    "extraction",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="products",
                        to="moonmining.extraction",
                    ),
                ),
                (
                    "ore_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="moonmining.eveoretype",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="extraction",
            name="refinery",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="extractions",
                to="moonmining.refinery",
            ),
        ),
        migrations.AddField(
            model_name="extraction",
            name="started_by",
            field=models.ForeignKey(
                default=None,
                help_text="Eve character who started this extraction",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AddConstraint(
            model_name="notification",
            constraint=models.UniqueConstraint(
                fields=("owner", "notification_id"), name="functional_pk_notification"
            ),
        ),
        migrations.AddConstraint(
            model_name="moonproduct",
            constraint=models.UniqueConstraint(
                fields=("moon", "ore_type"), name="functional_pk_moonproduct"
            ),
        ),
        migrations.AddConstraint(
            model_name="extractionproduct",
            constraint=models.UniqueConstraint(
                fields=("extraction", "ore_type"),
                name="functional_pk_extractionproduct",
            ),
        ),
        migrations.AddConstraint(
            model_name="extraction",
            constraint=models.UniqueConstraint(
                fields=("refinery", "started_at"), name="functional_pk_extraction"
            ),
        ),
    ]



import datetime as dt
from unittest.mock import patch

import pytz

from django.test import TestCase
from django.utils.timezone import now
from eveuniverse.models import EveMarketPrice, EveType

from app_utils.testing import NoSocketsTestCase

from ..models import EveOreType, Extraction, Moon, Refinery
from . import helpers
from .testdata.factories import ExtractionFactory, OwnerFactory, RefineryFactory
from .testdata.load_allianceauth import load_allianceauth
from .testdata.load_eveuniverse import load_eveuniverse
from .testdata.survey_data import fetch_survey_data

MANAGERS_PATH = "moonmining.managers"


class TestEveOreTypeManager(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    @patch(MANAGERS_PATH + ".MOONMINING_USE_REPROCESS_PRICING", False)
    def test_should_update_current_prices_with_market_price(self):
        # given
        ore_type = EveOreType.objects.get(name="Cinnabar")
        EveMarketPrice.objects.create(eve_type_id=ore_type.id, average_price=42)
        # when
        EveOreType.objects.update_current_prices()
        # then
        self.assertEqual(ore_type.extras.current_price, 42)

    @patch(MANAGERS_PATH + ".MOONMINING_REPROCESSING_YIELD", 0.7)
    @patch(MANAGERS_PATH + ".MOONMINING_USE_REPROCESS_PRICING", True)
    def test_should_update_current_prices_with_reprocessed_value(self):
        # given
        ore_type = EveOreType.objects.get(name="Cinnabar")
        tungsten = EveType.objects.get(id=16637)
        mercury = EveType.objects.get(id=16646)
        evaporite_deposits = EveType.objects.get(id=16635)
        EveMarketPrice.objects.create(eve_type=tungsten, average_price=7000)
        EveMarketPrice.objects.create(eve_type=mercury, average_price=9750)
        EveMarketPrice.objects.create(eve_type=evaporite_deposits, average_price=950)
        # when
        EveOreType.objects.update_current_prices()
        # then
        self.assertEqual(ore_type.extras.current_price, 4002.25)


class TestExtractionManager(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_eve_entities_from_allianceauth()

    def test_should_update_completed(self):
        # given
        refinery = RefineryFactory()
        extraction_1 = ExtractionFactory(
            refinery=refinery,
            started_at=dt.datetime(2021, 1, 1, 1, 0, tzinfo=pytz.UTC),
            chunk_arrival_at=dt.datetime(2021, 1, 1, 12, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2021, 1, 1, 15, 0, tzinfo=pytz.UTC),
            status=Extraction.Status.STARTED,
            create_products=False,
        )
        extraction_2 = ExtractionFactory(
            refinery=refinery,
            started_at=dt.datetime(2021, 1, 1, 2, 0, tzinfo=pytz.UTC),
            chunk_arrival_at=dt.datetime(2021, 1, 1, 15, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2021, 1, 1, 18, 0, tzinfo=pytz.UTC),
            status=Extraction.Status.STARTED,
            create_products=False,
        )
        extraction_3 = ExtractionFactory(
            refinery=refinery,
            started_at=dt.datetime(2021, 1, 1, 3, 0, tzinfo=pytz.UTC),
            chunk_arrival_at=dt.datetime(2021, 1, 1, 18, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2021, 1, 1, 21, 0, tzinfo=pytz.UTC),
            status=Extraction.Status.STARTED,
            create_products=False,
        )
        extraction_4 = ExtractionFactory(
            refinery=refinery,
            started_at=dt.datetime(2021, 1, 1, 4, 0, tzinfo=pytz.UTC),
            chunk_arrival_at=dt.datetime(2021, 1, 1, 4, 0, tzinfo=pytz.UTC),
            auto_fracture_at=dt.datetime(2021, 1, 1, 7, 0, tzinfo=pytz.UTC),
            status=Extraction.Status.CANCELED,
            create_products=False,
        )
        # when
        with patch(MANAGERS_PATH + ".now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 1, 1, 15, 30, tzinfo=pytz.UTC)
            Extraction.objects.all().update_status()
        # then
        extraction_1.refresh_from_db()
        self.assertEqual(extraction_1.status, Extraction.Status.COMPLETED)
        extraction_2.refresh_from_db()
        self.assertEqual(extraction_2.status, Extraction.Status.READY)
        extraction_3.refresh_from_db()
        self.assertEqual(extraction_3.status, Extraction.Status.STARTED)
        extraction_4.refresh_from_db()
        self.assertEqual(extraction_4.status, Extraction.Status.CANCELED)


class TestProcessSurveyInput(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        cls.user, cls.character_ownership = helpers.create_user_from_evecharacter(
            1001,
            permissions=[
                "moonmining.basic_access",
                "moonmining.extractions_access",
                "moonmining.add_refinery_owner",
            ],
            scopes=[
                "esi-industry.read_corporation_mining.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-corporations.read_structures.v1",
            ],
        )
        cls.survey_data = fetch_survey_data()

    @patch(MANAGERS_PATH + ".notify", new=lambda *args, **kwargs: None)
    def test_should_process_survey_normally(self):
        # when
        result = Moon.objects.update_moons_from_survey(
            self.survey_data.get(2), self.user
        )
        # then
        self.assertTrue(result)
        m1 = Moon.objects.get(pk=40161708)
        self.assertEqual(m1.products_updated_by, self.user)
        self.assertAlmostEqual(m1.products_updated_at, now(), delta=dt.timedelta(30))
        self.assertEqual(m1.products.count(), 4)
        self.assertEqual(m1.products.get(ore_type_id=45506).amount, 0.19)
        self.assertEqual(m1.products.get(ore_type_id=46676).amount, 0.23)
        self.assertEqual(m1.products.get(ore_type_id=46678).amount, 0.25)
        self.assertEqual(m1.products.get(ore_type_id=46689).amount, 0.33)

        m2 = Moon.objects.get(pk=40161709)
        self.assertEqual(m2.products_updated_by, self.user)
        self.assertAlmostEqual(m2.products_updated_at, now(), delta=dt.timedelta(30))
        self.assertEqual(m2.products.count(), 4)
        self.assertEqual(m2.products.get(ore_type_id=45492).amount, 0.27)
        self.assertEqual(m2.products.get(ore_type_id=45494).amount, 0.23)
        self.assertEqual(m2.products.get(ore_type_id=46676).amount, 0.21)
        self.assertEqual(m2.products.get(ore_type_id=46678).amount, 0.29)


class TestRefineryManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_allianceauth()
        helpers.generate_eve_entities_from_allianceauth()

    def test_should_return_ids(self):
        # given
        owner = OwnerFactory()
        RefineryFactory(id=1001, owner=owner)
        RefineryFactory(id=1002, owner=owner)
        # when
        result = Refinery.objects.ids()
        # then
        self.assertSetEqual(result, {1001, 1002})
